(* -- The example abstraction from the paper ----------------------------- *)

(* Do the abstraction, and apply the resulting abstracted rule. *)

(* Depending on your machine and memory, the following steps may
 * take wee while. Patience is a virtue. *)

(* What theorems do we depend on? *)
val All_deps = Inspect.dep_thms_not_in DStruct.thy stack_ref;
val PC_deps = Inspect.dep_thms_not_in DStruct.thy pc_thm;

(* Abstract StackRef, ArrayStack, and ListStack definitions *)
val Ref_deps = Inspect.dep_thms_in [StackRef.thy] pc_thm;
val R_thm = AbsTac.abs_thms Ref_deps pc_thm;
val Array_deps= Inspect.dep_thms_in [ArrayStack.thy] R_thm;
val A_thm = AbsTac.abs_thms Array_deps R_thm;
val List_deps= Inspect.dep_thms_in [ListStack.thy] A_thm;
val L_thm = AbsTac.abs_thms List_deps A_thm;

val rem_deps = Inspect.dep_thms_not_in DStruct.thy L_thm;

(* Function symbols *)
val dep_ops = Inspect.dep_ops_not_in DStruct.thy L_thm;

(* Anstracting over function symbols from StackRef *)
val R_ops = AbsTac.abs_ops_thy StackRef.thy L_thm;

(* Anstracting over function symbols from ArrayStack *)
val A_ops = AbsTac.abs_ops_thy ArrayStack.thy R_ops;

(* Anstracting over function symbols from ListStack *)
val L_ops = AbsTac.abs_ops_thy ListStack.thy A_ops;

(* Abstraction over types *)
val nuthy= theory_of_thm L_ops;
val absthm = Abstract.abstract_type nuthy (Type("IntDef.int", [])) 
                                          ("a", ["HOL.type"]) L_ops;

(* Now regain the original conclusion, i.e. c [= a. *)
val refdef_u = rewrite_rule [pc1_def, pc2_def, pc3_def] refines_def;
val giez_ref = refdef_u RS meta_eq_to_obj_eq RS iffD2;
val absrefthm = absthm RS giez_ref;


(* Now refine StringList: *)
use_thy "StringList";

val ref_str = read_instantiate_sg (sign_of StringList.thy) 
                                  [("a2", "StringList.sl")] absrefthm;
val ref2_str = read_instantiate_sg (sign_of StringList.thy) 
                                  [("lpop3", "StringList.stl"),
                                   ("lpush3", "StringList.scons"),
                                   ("ltop3", "StringList.shd"),
                                   ("lempty3", "StringList.strempty"),
                                   ("lops3", "StringList.sops")] ref_str;

val ref_str_to_array = foldl (fn (t, p)=> comp p t) (ref2_str, [sl_def, strempty_def, shd_def, sops_def, scons_def, stl_def]);

writeln "Abstraction successful."
