
structure Fibred_Sum_Cmd =
struct

exception FIBRED_SUM of string


fun repeatd delim scan = (scan -- Scan.repeat (OuterParse.$$$ delim |-- scan )) >> op :: || Scan.succeed []; 


val scan_map_param      = (OuterParse.$$$ "(" |-- (OuterParse.string --  
                                  (OuterParse.$$$ "|->" |-- (OuterParse.string --| OuterParse.$$$ ")"))));

val scan_map_param_list = (OuterParse.$$$ "[" |-- (repeatd "," scan_map_param) --| OuterParse.$$$ "]");

val scan_tmorph_list = (OuterParse.$$$ "[" |-- (repeatd "," OuterParse.name) --| OuterParse.$$$ "]");

val scan_thynm = OuterParse.name
val scan_obr = OuterParse.$$$ "(";
val scan_cbr = OuterParse.$$$ ")";
val scan_cmm = OuterParse.$$$ ",";
val scan_def_eq  = OuterParse.$$$ ":="
val scan_out = Scan.option(OuterParse.$$$ "with_output")
val scan_mLnm = OuterParse.name
val scan_mUnm = OuterParse.name; 
val scan_thym1nm = OuterParse.name;
val scan_thym2nm = OuterParse.name;

val scan_ren_key = OuterParse.$$$ "renames:"
val scan_renames = Scan.option(scan_ren_key |-- scan_map_param_list)

val fsum_scan = 
 scan_obr |-- (scan_thynm -- (scan_cmm |-- (scan_mLnm -- (scan_cmm |-- (scan_mUnm -- ( scan_cbr |-- (scan_def_eq |-- 
 (scan_tmorph_list -- (scan_renames -- scan_out))))))))) 


fun fsum_trans args =
   Toplevel.imperative (fn () => (let val (thynm, (mLnm, (mUnm, (tnms, (renames, out))))) = args
                                      fun get_tmorph nm = 
                                                 case (T_Toplevel.get_thymorph nm) of
                                                     Some t => 
                                                      (let val sg = ThyMorph.sig_morph t
                                                           val dom = SigMorph.src sg
                                                           val cod = SigMorph.trg sg
                                                           val domnm = Aux.name_of_thy dom
                                                           val codnm = Aux.name_of_thy cod
                                                           val t' = 
                                                             (let val tm = ThyMorph.cod_ext t [] (theory codnm)
                                                              in
                                                               ThyMorph.dom_ext tm [] [] (theory domnm)
                                                              end)
                                                       in
                                                        t'
                                                       end)
                                                   | None => raise FIBRED_SUM ("Unknown theory morphism " ^ nm)

                                      val tmorph_list = map get_tmorph tnms
                                      val ren = case renames of
                                                 Some r => r
                                               | None => []
                                      val outstream = case out of
                                                       None => None
                                                      | _ =>(Some (TextIO.openOut (thynm ^ ".thy")))

                                      val (rthy, tL, tU) =
                                        T_Toplevel.fibred_sum outstream ren (thynm, mLnm, mUnm) tmorph_list
                                       
                                      val _ = T_Toplevel.add_thymorph (mLnm, tL)
                                      val _ = T_Toplevel.add_thymorph (mUnm, tU)
                                      val sL = ThyMorph.sig_morph tL
                                      val sU = ThyMorph.sig_morph tU
                                      val src1 = Aux.name_of_thy (SigMorph.src sL)
                                      val src2 = Aux.name_of_thy (SigMorph.src sU)
                                      val trg1 = Aux.name_of_thy (SigMorph.trg sL)
                                      val trg2 = Aux.name_of_thy (SigMorph.trg sU)   
                                      val _ = writeln ("\nConstructed theory morphism " ^ 
                                        mLnm ^ " : " ^ src1 ^ " --> " ^ trg1 ^ "\n")
                                      val _ = writeln ("\nConstructed theory morphism " ^ 
                                        mUnm ^ " : " ^ src2 ^ " --> " ^ trg2 ^ "\n")
                                  in
                                   ()
                                  end))

fun fsum_parser tokens = (fsum_scan tokens) |>> (Toplevel.no_timing oo fsum_trans)

val fsum_cmd = OuterSyntax.improper_command "fibred_sum" "Calculate the colimit of the given list of theory morphisms with the same domain" OuterSyntax.Keyword.control fsum_parser;

val _ = OuterSyntax.add_keywords ["renames:", "with_output", ":="];
val _ = OuterSyntax.add_parsers [fsum_cmd];

end