
structure Parser_Ext =
struct



fun save_mixfix thynm (s1, s2, mxfx) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => T_Toplevel.add_mixfix (thynm ^ "." ^ s1, mxfx))
  in
   (s1, s2, mxfx)
  end 

fun save_mixfix2 thynm ((s1, s2, mxfx), s3) = 
  let val _ = (case mxfx of 
                 NoSyn => ()
               | _ => T_Toplevel.add_mixfix (thynm ^ "." ^ s1, mxfx))
  in
   ((s1, s2, mxfx), s3)
  end 


fun add_typedecls' (a, args, mx) thy =
  let val thynm = Aux.name_of_thy thy
      val _ = (case mx of
                 NoSyn => ()
               | _ => T_Toplevel.add_mixfix (thynm ^ "." ^ a, mx))
  in
   PureThy.add_typedecls [(a, args, mx)] thy
  end   

fun add_tyabbrs' abbrs thy =
  let val thynm = Aux.name_of_thy thy
      val _ = map (fn (a, args, T, mx) => (case mx of
                                            NoSyn => ()
                                           | _ => T_Toplevel.add_mixfix (thynm ^ "." ^ a, mx))) abbrs
  in
   Theory.add_tyabbrs abbrs thy
  end

(* *** A small extension of the Theory.add_consts function *** *) 
fun add_consts' cl thy =
  let val thynm = Aux.name_of_thy thy
  in
   Theory.add_consts (map (save_mixfix thynm) cl) thy
  end

(* *** A small extension of the IsarThy.add_constdefs function *** *) 
fun add_constdefs' cl thy =
  let val thynm = Aux.name_of_thy thy
  in
   IsarThy.add_constdefs (map (save_mixfix2 thynm) cl) thy
  end




val typedeclP =
  OuterSyntax.command "typedecl" "type declaration" OuterSyntax.Keyword.thy_decl
    (OuterParse.type_args -- OuterParse.name -- OuterParse.opt_infix >> 
     (fn ((args, a), mx) =>
      Toplevel.theory (add_typedecls' (a, args, mx))));


val typeabbrP =
  OuterSyntax.command "types" "declare type abbreviations" OuterSyntax.Keyword.thy_decl
    (Scan.repeat1
      (OuterParse.type_args -- OuterParse.name -- (OuterParse.$$$ "=" |-- 
       OuterParse.!!! (OuterParse.typ -- OuterParse.opt_infix')))
      >> (Toplevel.theory o add_tyabbrs' o
        map (fn ((args, a), (T, mx)) => (a, args, T, mx))));



val constsP =
  OuterSyntax.command "consts" "declare constants" OuterSyntax.Keyword.thy_decl
    (Scan.repeat1 OuterParse.const >> (Toplevel.theory o add_consts'));

val constdefsP =
  OuterSyntax.command "constdefs" "declare and define constants" OuterSyntax.Keyword.thy_decl
    (Scan.repeat1 (OuterParse.const -- OuterParse.term) 
    >> (Toplevel.theory o add_constdefs'));





(* *** Nonterminals *** *)

(* *** A small extension of the Theory.add_nonterminals function *** *) 
fun add_nonterminals' ntl thy =
  let val thynm = Aux.name_of_thy thy
      fun save_nt nt = 
         let val _ = T_Toplevel.add_nonterm (thynm, nt)
         in 
          nt
         end
  in
   Theory.add_nonterminals (map save_nt ntl) thy
  end

val nontermP =
  OuterSyntax.command "nonterminals" "declare types treated as grammar nonterminal symbols"
    OuterSyntax.Keyword.thy_decl (Scan.repeat1 OuterParse.name >> 
        (Toplevel.theory o add_nonterminals'));




(* *** syntax *** *)

(* *** A small extension of the Theory.add_modesyntax function *** *) 
fun add_modesyntax' ((s, b), sl) thy =
  let val thynm = Aux.name_of_thy thy
      val sg = sign_of thy
      fun save_syn sg (synnm, typestr, mxfx) = 
         let val typ = typ_of (read_ctyp sg typestr)
             val _ = T_Toplevel.add_syn_decl (thynm, (s, b, synnm, typ, mxfx))
         in 
          (synnm, typestr, mxfx)
         end
  in
   (uncurry Theory.add_modesyntax) ((s, b), map (save_syn sg) sl) thy
  end

val mode_spec =
  (OuterParse.$$$ "output" >> K ("", false)) || OuterParse.name -- Scan.optional (OuterParse.$$$ "output" >> K false) true;

val opt_mode = Scan.optional (OuterParse.$$$ "(" |-- OuterParse.!!! (mode_spec --| OuterParse.$$$ ")")) ("", true);

val syntaxP =
  OuterSyntax.command "syntax" "declare syntactic constants" OuterSyntax.Keyword.thy_decl
    (opt_mode -- Scan.repeat1 OuterParse.const >> (Toplevel.theory o add_modesyntax'));




(* *** translations *** *)

fun add_trrules' trl thy =
 let val thyname = Aux.name_of_thy thy
     fun save_trrule tr =
         let val _ = T_Toplevel.add_trrule (thyname, tr)
         in
          tr
         end
 in
  Theory.add_trrules (map save_trrule trl) thy
 end  

val trans_pat =
  Scan.optional (OuterParse.$$$ "(" |-- OuterParse.!!! (OuterParse.xname --| OuterParse.$$$ ")")) "logic" -- OuterParse.string;

fun trans_arrow toks =
  ((OuterParse.$$$ "\\<rightharpoonup>" || OuterParse.$$$ "=>") >> K Syntax.ParseRule ||
    (OuterParse.$$$ "\\<leftharpoondown>" || OuterParse.$$$ "<=") >> K Syntax.PrintRule ||
    (OuterParse.$$$ "\\<rightleftharpoons>" || OuterParse.$$$ "==") >> K Syntax.ParsePrintRule) toks;

val trans_line =
  trans_pat -- OuterParse.!!! (trans_arrow -- trans_pat)
    >> (fn (left, (arr, right)) => arr (left, right));

val translationsP =
  OuterSyntax.command "translations" "declare syntax translation rules" OuterSyntax.Keyword.thy_decl
    (Scan.repeat1 trans_line >> (Toplevel.theory o add_trrules'));




(* *** Translation functions *** *)

fun parse_ast_translation' s thy =
 let val thyname = Aux.name_of_thy thy
     val _ = T_Toplevel.add_parse_ast_trans (thyname, s)
 in
  IsarThy.parse_ast_translation s thy
 end

val parse_ast_translationP =
  OuterSyntax.command "parse_ast_translation" "install parse ast translation functions" OuterSyntax.Keyword.thy_decl
    (OuterParse.text >> (Toplevel.theory o parse_ast_translation'));



fun parse_translation' s thy =
 let val thyname = Aux.name_of_thy thy
     val _ = T_Toplevel.add_parse_trans (thyname, s)
 in
  IsarThy.parse_translation s thy
 end

val parse_translationP =
  OuterSyntax.command "parse_translation" "install parse translation functions" OuterSyntax.Keyword.thy_decl
    (OuterParse.text >> (Toplevel.theory o parse_translation'));



fun print_translation' s thy =
 let val thyname = Aux.name_of_thy thy
     val _ = T_Toplevel.add_print_trans (thyname, s)
 in
  IsarThy.print_translation s thy
 end

val print_translationP =
  OuterSyntax.command "print_translation" "install print translation functions" OuterSyntax.Keyword.thy_decl
    (OuterParse.text >> (Toplevel.theory o print_translation'));





fun print_ast_translation' s thy =
 let val thyname = Aux.name_of_thy thy
     val _ = T_Toplevel.add_print_ast_trans (thyname, s)
 in
  IsarThy.print_ast_translation s thy
 end

val print_ast_translationP =
  OuterSyntax.command "print_ast_translation" "install print ast translation functions" OuterSyntax.Keyword.thy_decl
    (OuterParse.text >> (Toplevel.theory o print_ast_translation'));




fun token_translation' s thy =
 let val thyname = Aux.name_of_thy thy
     val _ = T_Toplevel.add_token_trans (thyname, s)
 in
  IsarThy.token_translation s thy
 end

val token_translationP =
  OuterSyntax.command "token_translation" "install token translation functions" OuterSyntax.Keyword.thy_decl
    (OuterParse.text >> (Toplevel.theory o token_translation'));


fun theory' (s, sl, l) t =
  let val t' = IsarThy.theory (s, sl, l) t
      val _ = T_Toplevel.upd_rem_tabs s
  in
   t'
  end

val theoryP =
  OuterSyntax.command "theory" "begin theory" OuterSyntax.Keyword.thy_begin
    (ThyHeader.args >> (Toplevel.print oo theory'));



val _ = OuterSyntax.add_parsers [typeabbrP, constsP, constdefsP, nontermP, syntaxP, translationsP];
val _ = OuterSyntax.add_parsers [parse_ast_translationP, parse_translationP, 
                                 print_translationP, print_ast_translationP,
                                 token_translationP]
val _ = OuterSyntax.add_parsers [theoryP]


end