(*
Adds the 'sigmorph' command to Isabelle/Isar's outer
syntax.
*)

structure SigMorph_cmd =
struct

(* ------------------------------------------------------------------------- *)
(* common functions for argument parsing/scanning                            *)
(* ------------------------------------------------------------------------- *)

(* ------------------------------------------------------------------------- *)
(* the arguments of 'sigmorph' are of   *)
(* the form unique_sgmph_name : src_thy_name --> trg_thy_name 
            maps [("typedecl_name1" |-> "type1"), ...]
                 [("opname1" |-> "opname2"),...]   
            <strictly>              *)
(* ------------------------------------------------------------------------- *)




        fun repeatd delim scan = (scan -- Scan.repeat (OuterParse.$$$ delim |-- scan ) >> op ::) || Scan.succeed []  


        val scan_map_param      = (OuterParse.$$$ "(" |-- (OuterParse.string --  
                                  (OuterParse.$$$ "|->" |-- (OuterParse.string --| OuterParse.$$$ ")"))))

        val scan_map_param_list = (OuterParse.$$$ "[" |-- (repeatd "," scan_map_param) 
                                                      --| OuterParse.$$$ "]")

            


(* ------------------------------------------------------------------------- *)
(* set up the 'sigmorph' command                                               *)
(* ------------------------------------------------------------------------- *)


val scan_col  = OuterParse.$$$ ":"
val scan_arrow = OuterParse.$$$ "-->"
val scan_maps = OuterParse.$$$ "maps"
val scan_strictly = OuterParse.$$$ "strictly"
val scan_tpmp = Scan.option(scan_maps |-- scan_map_param_list)
val scan_opmp = Scan.option(scan_map_param_list)
val scan_mrphnm =OuterParse.name 
val scan_srcnm = OuterParse.name
val scan_trgnm = OuterParse.name




        val sigmorph_scan_args = 
            scan_mrphnm -- (scan_col |-- (scan_srcnm -- (scan_arrow |-- (scan_trgnm --  
            (scan_tpmp -- (scan_opmp -- (Scan.option scan_strictly)))))))
                           

(* ------------------------------------------------------------------------- *)
(* the 'sigmorph' command is mapped to 'SigMorph.make' or 'T_Toplevel.try_sigmorph'  *)
(* ------------------------------------------------------------------------- *)
 

  fun create_typemap sg arg = 
    let val ltp = Aux.get_ltypes sg
        fun f (t1, t2) = (if (t2 mem ltp) then (t1, SigMorph.Simple t2)
                          else (t1, (SigMorph.Comp o typ_of o (read_ctyp sg)) t2))
    in
         (map f arg) : (SigMorph.type_map SigMorph.map_rep)              
    end


  fun sigmorph_trans args
          =
              Toplevel.theory
                ( fn thisthy =>
                      (let val (mrphnm, (srcnm, (trgnm, (tpmp, (opmp, opt))))) = args;
                           val src = if (Aux.name_of_thy thisthy = srcnm) then thisthy else (theory srcnm)
                           val trg = if (Aux.name_of_thy thisthy = trgnm) then thisthy else (theory trgnm)

                           val (tpMap, opMap) = (case tpmp of
                                                  Some t => (let val sg = sign_of trg
                                                                 val t' = create_typemap sg t      
                                                             in
                                                              case opmp of
                                                                Some m => (t', m)
                                                               | None => (t', [])
                                                             end)
                                                 | None => ([], []))
                               
                           val sigma = (case opt of
                                         Some s => SigMorph.make src trg tpMap opMap
                                        | None => T_Toplevel.try_sigmorph src trg tpMap opMap) 
                           val _ = writeln ("\nConstructed signature morphism " ^ mrphnm ^ " : " ^ srcnm ^ 
                                            " --> " ^ trgnm ^ "\n")
                           val _ = T_Toplevel.add_sigmorph (mrphnm, sigma)
                       in 
                         thisthy
                       end)
                 )



        fun sigmorph_parser tokens = (sigmorph_scan_args tokens) |>> sigmorph_trans 
                                        

 val sigmorph_cmd = OuterSyntax.command "sigmorph" "Construct a signature morphism between two theories by given type and operation maps" OuterSyntax.Keyword.thy_goal sigmorph_parser;


val _ = OuterSyntax.add_keywords["-->"];
val _ = OuterSyntax.add_keywords["maps"];
val _ = OuterSyntax.add_keywords["strictly"]; 
val _ = OuterSyntax.add_keywords["|->"]; 

val _ = OuterSyntax.add_parsers [sigmorph_cmd];

end;
