(*
Adds the 't_compose' command to Isabelle/Isar's outer
syntax.
*)

structure T_Compose =
struct

exception T_COMPOSE of string

(* ------------------------------------------------------------------------- *)
(* set up the 't_compose' command                                         *)
(* ------------------------------------------------------------------------- *)



val scan_comp  = OuterParse.$$$ "o"
val scan_tmrphnm1 = OuterParse.name; 
val scan_tmrphnm2 = OuterParse.name;
val scan_ctmrphnm = OuterParse.name;
val scan_def_eq  = OuterParse.$$$ ":="

val t_compose_scan = scan_ctmrphnm -- (scan_def_eq |-- (scan_tmrphnm2 -- (scan_comp |-- scan_tmrphnm1)))


fun t_compose_trans args =
  Toplevel.theory (fn thy =>
  ( let val (ctnm, (tnm2, tnm1)) = args 
        val tm1 = case (T_Toplevel.get_thymorph tnm1) of
                    Some t => t
                  | None   => raise T_COMPOSE ("Unknown theory morphism " ^ tnm1)
        val tm2 = case (T_Toplevel.get_thymorph tnm2) of
                    Some t => t
                  | None   => raise T_COMPOSE ("Unknown theory morphism " ^ tnm2)
   
        val sigma1 = ThyMorph.sig_morph tm1
        val dom    = SigMorph.src sigma1
        val domnm  = Aux.name_of_thy dom
        val cod1   = SigMorph.trg sigma1
        val cod1nm = Aux.name_of_thy cod1

        val sigma2 = ThyMorph.sig_morph tm2
        val dom2   = SigMorph.src sigma2
        val dom2nm = Aux.name_of_thy dom2  
        val cod    = SigMorph.trg sigma2
        val codnm = Aux.name_of_thy cod

        val thisnm  = Aux.name_of_thy thy

        val tm1' = if (thisnm = domnm) (* i.e. the dom is an Isar transition
                                          so we have to update the sigma's domain *) 
                   then (ThyMorph.dom_ext tm1 [] [] thy)   
                   else (ThyMorph.dom_ext tm1 [] [] (theory domnm))

        val tm1'' = if (thisnm = cod1nm) then (ThyMorph.cod_ext tm1' [] thy)   
                                         else (ThyMorph.cod_ext tm1' [] (theory cod1nm))

        val tm2' = if (thisnm = dom2nm) then (ThyMorph.dom_ext tm2 [] [] thy)
                                        else (ThyMorph.dom_ext tm2 [] [] (theory dom2nm))

        val tm2'' = if (thisnm = codnm) then (ThyMorph.cod_ext tm2' [] thy)
                                        else (ThyMorph.cod_ext tm2' [] (theory codnm))

        val ctm = T_Toplevel.thymorph_compose tm1'' tm2''
        val sigma' = ThyMorph.sig_morph ctm
        val ntrg   = SigMorph.trg sigma' 
   
        val _ = writeln ("\nConstructed theory morphism " ^ ctnm ^ " : " ^ domnm ^ " --> " ^ codnm) 
        val _ = T_Toplevel.add_thymorph (ctnm, ctm)
    in
      if (Aux.name_of_thy ntrg = thisnm) then ntrg else thy
    end))


fun t_compose_parser tokens = (t_compose_scan tokens) |>> t_compose_trans


val t_compose_cmd = OuterSyntax.command "t_compose" "Declare a theory morphism as the composition of theory morphisms" 
                         OuterSyntax.Keyword.thy_goal t_compose_parser;


val _ = OuterSyntax.add_keywords["o"]; 
val _ = OuterSyntax.add_parsers [t_compose_cmd];

end;
