(*
Adds the 't_print' command to Isabelle/Isar's outer
syntax.
*)

structure T_Print =
struct

exception T_PRINT of string;

val t_print_arg = OuterParse.name;


fun t_print_trans arg
          =
              Toplevel.imperative
                ( fn () =>
                      (let val res = if arg = "sigmorphs" then T_Toplevel.print_sigmorphs ()
                                     else if arg = "thymorphs" then T_Toplevel.print_thymorphs ()
                                     (* else if arg = "trafos" then T_Toplevel.print_trafos () *)
                                     else raise T_PRINT ("Unknown argument " ^ arg)
                             
                       in
                          writeln res  
                       end)
                 ) 

fun t_print_parser tokens = (t_print_arg tokens) |>> t_print_trans

val t_print_cmd = OuterSyntax.command "t_print" "Shows the list of current: sigmorphs -- signature morphisms, thymorphs -- theory morphisms, trafos -- tranformation rules" OuterSyntax.Keyword.control t_print_parser;
 

val _ = OuterSyntax.add_parsers [t_print_cmd];

end;
                                        