(*
Adds the 'thymorph' command to Isabelle/Isar's outer
syntax.
*)

structure ThyMorph_cmd =
struct

exception THYMORPH_CMD of string

(* ------------------------------------------------------------------------- *)
(* common functions for argument parsing/scanning                            *)
(* ------------------------------------------------------------------------- *)

(* ------------------------------------------------------------------------- *)
(* the arguments of 'sigmorph' are of   *)
(* the form unique_thymph_name : src_thy_name --> trg_thy_name 
            maps [("typedecl_name1" |-> "type1"), ...]
                 [("opname1" |-> "opname2"),...]
                 [("thmname1" |-> "thmname2"),...]     
            <strictly>
   or without 'maps'                                                         *)
(* ------------------------------------------------------------------------- *)




        fun repeatd delim scan = (scan -- Scan.repeat (OuterParse.$$$ delim |-- scan ) >> op ::) || Scan.succeed []  


        val scan_map_param      = (OuterParse.$$$ "(" |-- (OuterParse.string --  
                                  (OuterParse.$$$ "|->" |-- (OuterParse.string --| OuterParse.$$$ ")"))))

        val scan_map_param_list = (OuterParse.$$$ "[" |-- (repeatd "," scan_map_param) 
                                                      --| OuterParse.$$$ "]")

            


(* ------------------------------------------------------------------------- *)
(* set up the 'sigmorph' command                                               *)
(* ------------------------------------------------------------------------- *)


val scan_col  = OuterParse.$$$ ":"
val scan_arrow = OuterParse.$$$ "-->"
val scan_maps = OuterParse.$$$ "maps"
val scan_optmaps = Scan.option(scan_maps)
val scan_strictly = OuterParse.$$$ "strictly"
val scan_sigma = OuterParse.$$$ "by_sigmorph"
val scan_tpmp = Scan.option(scan_maps |-- scan_map_param_list);
val scan_opmp = Scan.option(scan_map_param_list);
val scan_thmmp = Scan.option(scan_map_param_list);
val scan_mrphnm =OuterParse.name; 
val scan_srcnm = OuterParse.name;
val scan_trgnm = OuterParse.name;


fun lift1 f t =
    let val ((mrphnm, (srcnm, (trgnm, (tpmp, (opmp, (thmmp, opt)))))), t') = f t
    in 
      ((mrphnm, None, srcnm, trgnm, tpmp, opmp, thmmp, opt), t')
    end
fun lift2 f t =
    let val ((mrphnm, (sigma_nm, thmmp)), t') = f t
    in 
      ((mrphnm, Some sigma_nm, "", "", None, None, thmmp, None), t')
    end 


val thymorph_scan_args = 
     (lift1 (scan_mrphnm -- 
            (scan_col |-- (scan_srcnm -- (scan_arrow |-- (scan_trgnm --  
            (scan_tpmp -- (scan_opmp -- (scan_thmmp -- (Scan.option scan_strictly))))))))))
        ||
     (lift2 (scan_mrphnm -- (scan_sigma |-- (scan_mrphnm -- (Scan.option (scan_maps |-- scan_map_param_list))))))                           



  fun create_typemap sg arg = 
    let val ltp = Aux.get_ltypes sg
        fun f (t1, t2) = (if (t2 mem ltp) then (t1, SigMorph.Simple t2)
                          else (t1, (SigMorph.Comp o typ_of o (read_ctyp sg)) t2))
    in
         (map f arg) : (SigMorph.type_map SigMorph.map_rep)              
    end


  fun thymorph_trans args =
   Toplevel.theory
    ( fn thisthy =>
            (let val (mrphnm, opt_sigma, srcnm, trgnm, tpmp, opmp, thmmp, opt) = args
             in
               case opt_sigma of
                None =>
                 (let val src = if (Aux.name_of_thy thisthy = srcnm) then thisthy else (theory srcnm)
                      val trg = if (Aux.name_of_thy thisthy = trgnm) then thisthy else (theory trgnm)

                      val (tpMap, opMap, thmMap) = (case tpmp of 
                                                      Some t => (let val sg = sign_of trg
                                                                     val t' = create_typemap sg t
                                                                 in
                                                                   (case opmp of
                                                                     Some omp => (case thmmp of
                                                                                   Some tmp => (t', omp, tmp)
                                                                                 | None => (t', omp, []))
                                                                   | Npne => (t', [], []))
                                                                 end)
                                                    | None => ([], [], []))
                                                     
                       val sigma = (case opt of
                                      Some s => SigMorph.make src trg tpMap opMap
                                    | None => T_Toplevel.try_sigmorph src trg tpMap opMap)
                       val tmorph = (case opt of
                                       Some s => ThyMorph.make sigma thmMap
                                     | None => T_Toplevel.try_thymorph sigma)    
 
                       val _ = writeln ("\nConstructed theory morphism " ^ 
                                        mrphnm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n")
                       val _ = T_Toplevel.add_thymorph (mrphnm, tmorph)
                   in 
                    thisthy
                   end)

                | Some sigma_nm =>
                  (let val sigma = case (T_Toplevel.get_sigmorph sigma_nm) of
                                    Some s => s
                                   | None  => raise THYMORPH_CMD ("Unknown signature morphism " ^ sigma_nm)
                       val src   = SigMorph.src sigma
                       val trg   = SigMorph.trg sigma
                       val srcnm = Aux.name_of_thy src
                       val trgnm = Aux.name_of_thy trg
                       val thynm = Aux.name_of_thy thisthy              
                       val sigma' = if (thynm = srcnm) (* i.e. the src is an Isar transition
                                                          so we have to update the sigma's domain *) 
                                    then (SigMorph.dom_ext sigma [] [] thisthy)  
                                    else sigma

                       val sigma'' = if (thynm = trgnm) (* i.e. the trg is an Isar transition
                                                           so we have to update the sigma's codomain *) 
                                      then (SigMorph.cod_ext sigma' thisthy)  
                                      else sigma'
                       val thmMap = case thmmp of
                                     Some tmp => tmp
                                    | None => []
                       val tmorph = (case thmMap of
                                      [] => T_Toplevel.try_thymorph sigma''
                                     | _ => ThyMorph.make sigma'' thmMap)
                       val _ = writeln ("\nConstructed theory morphism " ^ 
                                        mrphnm ^ " : " ^ srcnm ^ " --> " ^ trgnm ^ "\n")
                       val _ = T_Toplevel.add_thymorph (mrphnm, tmorph)
                   in 
                    thisthy
                   end)
             end)   )


fun thymorph_parser tokens = (thymorph_scan_args tokens) |>> thymorph_trans 
                                        

 val thymorph_cmd = OuterSyntax.command "thymorph" "Construct a theory morphism between two theories by given type, operation and theorem maps or by given signature morphism and theorem map" OuterSyntax.Keyword.thy_goal thymorph_parser;


val _ = OuterSyntax.add_keywords["-->"];
val _ = OuterSyntax.add_keywords["maps"];
val _ = OuterSyntax.add_keywords["strictly"]; 
val _ = OuterSyntax.add_keywords["by_sigmorph"];
val _ = OuterSyntax.add_keywords["|->"]; 

val _ = OuterSyntax.add_parsers [thymorph_cmd];

end;
