(*
Adds the 'translate_thm' command to Isabelle/Isar's outer
syntax.
*)

structure Translate_Thm =
struct

exception TRANSLATE_THM of string

(* ------------------------------------------------------------------------- *)
(* set up the 'translate_thm' command                                         *)
(* ------------------------------------------------------------------------- *)




val scan_tmrphnm = OuterParse.name; 
val scan_thmnm = OuterParse.string;
val scan_thmrnm = OuterParse.name;
val scan_along  = OuterParse.$$$ "along"
val scan_as  = OuterParse.$$$ "as"

val translate_thm_scan = scan_thmnm -- (scan_as |-- (scan_thmrnm -- (scan_along |-- scan_tmrphnm)))

fun translate_thm_trans args =
  Toplevel.theory (fn thy =>
  ( let val (thmnm, (thmrnm, tmrphnm)) = args 
        val tm = case (T_Toplevel.get_thymorph tmrphnm) of
                    Some t => t
                  | None   => raise TRANSLATE_THM ("Unknown theory morphism " ^ tmrphnm)
  
        val sigma = ThyMorph.sig_morph tm
        val dom    = SigMorph.src sigma
        val domnm  = Aux.name_of_thy dom
        val cod   = SigMorph.trg sigma
        val codnm = Aux.name_of_thy cod

        val thisnm  = Aux.name_of_thy thy

        val tm' = if (thisnm = domnm) (* i.e. the dom is an Isar transition
                                          so we have to update the sigma's domain *) 
                   then (ThyMorph.dom_ext tm [] [] thy)   
                   else  (ThyMorph.dom_ext tm [] [] (theory domnm)) 

        val tm'' = if (thisnm = codnm) then (ThyMorph.cod_ext tm' [] thy)   
                                         else (ThyMorph.cod_ext tm' [] (theory codnm))  

        val tm''' = T_Toplevel.translate_thm tm'' thmnm thmrnm
        val _     = T_Toplevel.add_thymorph (tmrphnm, tm''')
        val sigma = ThyMorph.sig_morph tm'''
        val last_trg = SigMorph.trg sigma
    in
     if ((Aux.name_of_thy last_trg) = thisnm) then last_trg else thy
    end))



fun translate_thm_parser tokens = translate_thm_scan tokens |>> translate_thm_trans

val translate_thm_cmd = OuterSyntax.command "translate_thm" "Translate a theorem along a theory morphism" 
                         OuterSyntax.Keyword.thy_goal translate_thm_parser;

val _ = OuterSyntax.add_keywords["as"]; 
val _ = OuterSyntax.add_keywords["along"];
val _ = OuterSyntax.add_parsers [translate_thm_cmd];

end;
