
header {* Greedy-Algorithm *}

theory Greedy = Greedy_Aux3 :


section {* The parameters of Greedy *}

typedecl D;
typedecl V;

consts
U :: "D list"
val :: "D \<Rightarrow> V"
V_op :: "[V, V] \<Rightarrow> V"          (infixl "+++" 65)
V_ord :: "[V, V] \<Rightarrow> bool"      (infixl "<<" 45)
V_unit :: V

subsection {* U properties and sorting *}

constdefs
D_ord :: "[D, D] \<Rightarrow> bool "    (infixl "<<*" 45)
"D_ord \<equiv> \<lambda> x y. val x << val y"

consts
sorting_function :: "D list \<Rightarrow> D list"
axioms
 A_sort : " \<forall> l. sorted op <<* (sorting_function l) \<and> l <~~> sorting_function l "


axioms
 U1 : " distinct U "
 U2 : " sorted op <<* U "


subsection {* Order properties *}

axioms
 V1 : " \<forall> x. V_unit +++ x = x "
 V2 : " \<forall> x y. x +++ y = y +++ x "
 V3 : " \<forall> x y z. x +++ (y +++ z) = x +++ y +++ z"

axioms
 V_ord_refl  : " \<forall> x. x << x"
 V_ord_trans : " \<forall> x y z. x << y \<and> y << z \<longrightarrow> x << z "
 V_ord_mono  : " \<forall> x y d. x << y \<longrightarrow> (x +++ d) << (y +++ d) " 
 V_ord_total : " \<forall> x y. x << y \<or> x = y \<or> y << x "



subsection {* Feasible solutions *}

consts
Feasible :: "D list \<Rightarrow> bool"

axioms
 FA1 : "\<forall> A. Feasible A \<longrightarrow> (sublist A U) \<and> (distinct A) "
 FA2 : "\<forall> A B. Feasible A \<and> A <~~> B \<longrightarrow> Feasible B "


subsection {* Matroid properties *}

axioms
 M1 : " Feasible []"
 M2 : " \<forall> A B. distinct A \<and> Feasible B \<and> sublist A B \<longrightarrow> Feasible A "
 M3 : " \<forall> A B. Feasible A \<and> Feasible B \<and> length A < length B 
                       \<longrightarrow> (\<exists> e. (e mem B) \<and> (\<not> (e mem A)) 
                                          \<and> (Feasible (A @ [e]))) "






section {* Derived properties *}



subsection {* Basic properties of Feasible *}

lemma Feas1 : " \<lbrakk> Feasible A; e mem A \<rbrakk> \<Longrightarrow> e mem U "
apply(insert FA1);
apply(drule_tac x="A" in spec);
apply(drule mp, simp);
by (simp add: sublist_def);


lemma Feas_app : " Feasible (xs @ ys) \<Longrightarrow> Feasible xs \<and> Feasible ys "
apply(rule conjI);
apply(rule M2[rule_format]);
apply(rule conjI);
apply(drule FA1[rule_format]);
apply(simp);
apply(rule conjI);
apply(assumption);
apply(simp add: sublist_def);
apply(clarify);
apply(simp add: set_mem_eq);
apply(rule M2[rule_format]);
apply(rule conjI);
apply(drule FA1[rule_format]);
apply(simp);
apply(rule conjI);
apply(assumption);
apply(simp add: sublist_def);
apply(clarify);
by(simp add: set_mem_eq);


lemma Feas_neg1 : " \<not> Feasible xs \<Longrightarrow> \<forall> ys. \<not> Feasible (ys @ xs) "  
apply(clarify);
apply(erule notE);
apply(drule Feas_app);
by simp;


lemma Feas_neg2 : " \<not> Feasible (P @ [x]) \<Longrightarrow> \<forall> ys. \<not> Feasible (P @ ys @ [x]) "  
apply(clarify);
apply(erule notE);
apply(rule M2[rule_format]);
apply(rule conjI);
apply(drule FA1[rule_format]);
apply(simp);
apply(rule conjI);
apply(assumption);
apply(rule sublist_subset[THEN iffD2]);
apply(rule subsetI);
apply(simp);
apply(erule disjE);
by simp_all;

lemma Feasible_length : " Feasible A \<Longrightarrow> length A \<le> length U "
apply(drule FA1[rule_format]);
apply(erule conjE);
apply(rule not_not[THEN iffD1]);
apply(rule notI);
apply(drule not_leE);
apply(drule sublist_length1);
apply(assumption);
apply(erule notE);
by assumption;




subsection {* Definition and properties of greedy *}

consts
greedy :: "D list \<Rightarrow> D list \<Rightarrow> D list \<Rightarrow> (D list \<times> D list) "
primrec 
"greedy P N [] = (P, N)"
"greedy P N (x#xs) = (if (Feasible (P @ [x])) then (greedy (P @ [x]) N xs) else (greedy P (x#N) xs))" 



lemma greedy0[rule_format] : " \<forall> P. Feasible (P @ xs) \<longrightarrow> (fst (greedy P N xs)) = P @ xs "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(drule_tac x="P @ [a]" in spec);
apply(drule mp);
apply(simp);
apply(simp);
apply(rule impI);
apply(erule notE);
apply(rule_tac ys1=list in Feas_app[THEN conjunct1]);
apply(simp);
done


lemma greedy1[simplified] : " Feasible ([] @ xs) \<Longrightarrow> (fst (greedy [] N xs)) = [] @ xs "
apply(rule greedy0);
by simp;


theorem greedy2 : " \<forall> xs P N. Feasible P \<longrightarrow> Feasible (fst (greedy P N xs)) "
apply(rule allI);
apply(induct_tac xs);
by (simp_all);


corollary greedy3 : " \<forall> N xs. Feasible (fst (greedy [] N xs)) "
apply(clarify);
apply(rule greedy2[rule_format]);
by (rule M1);

theorem greedy4 : "\<forall> xs P N. (xs @ P @ N) <~~> (fst (greedy P N xs) @ snd (greedy P N xs)) "
apply(rule allI);
apply(induct_tac xs);
apply(simp);
apply(simp);
apply(intro allI);
apply(rule conjI);
apply(rule impI);
apply(drule_tac x="P @ [a]" in spec);
apply(drule_tac x="N" in spec);
apply(rule Permutation.trans);
prefer 2;
apply(assumption);
apply(simp);
apply(rule Permutation.trans);
apply(rule perm_append_Cons);
apply(simp);
apply(rule Permutation.trans);
apply(rule perm_append_Cons);
apply(simp);
apply(rule impI);
apply(drule_tac x="P" in spec);
apply(drule_tac x="a # N" in spec);
apply(rule Permutation.trans);
prefer 2;
apply(assumption);
apply(rule Permutation.trans);
apply(rule perm_append_Cons);
apply(simp);
apply(rule Permutation.trans);
apply(rule perm_append_Cons);
by(simp);



theorem greedy_union : "\<forall> x xs P N. x mem xs 
                        \<longrightarrow> (x mem (fst (greedy P N xs)) \<or> x mem (snd (greedy P N xs)))"
apply(clarify);
apply(insert greedy4);
apply(drule_tac x="xs" in spec);
apply(drule_tac x="P" in spec);
apply(drule_tac x="N" in spec);
apply(drule_tac xs="xs @ P @ N" in perm_mem);
apply(simp add: set_mem_eq);
apply(rule disjI1);
apply(assumption);
by(simp add: set_mem_eq);


corollary greedy_U_union : " x mem U \<Longrightarrow> 
                             (x mem (fst (greedy [] [] U)) \<or> x mem (snd (greedy [] [] U))) "
apply(drule greedy_union[rule_format]);
apply(erule disjE);              
apply(rule disjI1);
apply(assumption);
apply(rule disjI2);
apply(assumption);
done


theorem greedy5 : "\<forall> P N. x mem (snd (greedy P N xs)) \<and> (distinct P) \<and> (distinct xs) \<and>
                                (\<forall> y. y mem xs \<longrightarrow> (\<not> (y mem P))) \<and>
                                (\<forall> y. y mem N \<longrightarrow> (\<not> (Feasible (P @ [y]))))
                 \<longrightarrow> 
                 (\<exists> ys. (distinct ys) \<and> 
                                (sublist ys (fst (greedy P N xs))) \<and> (\<not> (Feasible (ys @ [x])))) " 
apply(induct_tac xs);
apply(clarify);
apply(rule_tac x="P" in exI);
apply(drule_tac x="x" in spec);
back;
apply(drule mp, simp);
apply(simp add: sublist_def);
apply(clarify);
apply(simp);
apply(rule conjI);
apply(rule impI);
apply(split split_if_asm);
apply(drule_tac x="P @ [a]" in spec);
apply(drule_tac x="N" in spec);
apply(drule mp);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(simp);
apply(drule_tac x=a in spec);
apply(drule mp, simp);
apply(simp add: set_mem_eq);
apply(rule conjI);
apply(clarify);
apply(case_tac "a = y");
apply(simp add: set_mem_eq);
apply(drule_tac x="y" in spec);
apply(drule mp, simp);
apply(simp add: set_mem_eq);
apply(clarify);
apply(drule_tac x="y" in spec);
back;
apply(drule mp, assumption);
apply(drule Feas_neg2);
apply(drule_tac x="[a]" in spec);
apply(erule notE);
back;
apply(simp);
apply(assumption);
apply(erule notE);
apply(assumption);
apply(rule impI);
apply(split split_if_asm);
apply(erule notE);
apply(assumption);
apply(drule_tac x="P" in spec);
apply(drule_tac x="a # N" in spec);
apply(drule mp);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(clarify);
apply(drule_tac x=y in spec);
apply(drule mp, simp);
apply(simp);
apply(clarify);
apply(erule notE);
apply(case_tac "y = a");
apply(simp);
apply(drule_tac x="y" in spec);
back;
apply(drule mp);
apply(simp add: set_mem_eq);
apply(erule notE, assumption);
by (assumption);


corollary greedy5_U : "x mem (snd (greedy [] [] U)) \<Longrightarrow> 
                        (\<exists> ys. (distinct ys) \<and> 
                                       (sublist ys (fst (greedy [] [] U))) \<and> (\<not> (Feasible (ys @ [x])))) " 
apply(rule greedy5[rule_format]);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(rule U1);
by simp;


theorem greedy_not_exch:
             " e mem (snd (greedy [] [] U)) 
               \<Longrightarrow> \<not> (Feasible (fst (greedy [] [] U) @ [e]))" 
apply(drule greedy5_U);
apply(clarify);
apply(erule notE);
apply(rule M2[rule_format]);
apply(rule conjI);
apply(drule FA1[rule_format]);
apply(simp);
apply(elim conjE);
apply(rule notI);
apply(erule notE);
apply(drule sublist_subset[THEN iffD1]);
apply(erule subsetD);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(drule sublist_subset[THEN iffD1]);
apply(rule sublist_subset[THEN iffD2]);
apply(rule subsetI);
apply(simp);
apply(erule disjE);
apply(simp);
apply(rule disjI2);
apply(erule subsetD);
apply(assumption);
done


subsection {* max_range property *}

constdefs
max_range :: "D list \<Rightarrow> bool"
"max_range A \<equiv> (Feasible A) \<and> (\<forall> B. Feasible B \<longrightarrow> length B \<le> length A) " 


theorem greedy_max_range : " max_range (fst (greedy [] [] U)) "
apply(simp add: max_range_def);
apply(insert greedy3);
apply(drule_tac x="[]" in spec);
apply(drule_tac x="U" in spec);
apply(rule conjI);
apply(assumption);
apply(clarify);
apply(rule not_not[THEN iffD1]);
apply(rule notI);
apply(drule not_leE);
apply(frule FA1[rule_format]);
apply(frule FA1[rule_format]);
back;
apply(clarify);
apply(drule M3[rule_format, OF conjI]);
apply(rule conjI);
apply(assumption);
apply(assumption);
apply(clarify);
apply(subgoal_tac "e mem U");
prefer 2;
apply(simp add: sublist_def);
apply(subgoal_tac "e mem (snd (greedy [] [] U))");
prefer 2;
apply(drule greedy_U_union);
apply(erule disjE);
apply(erule notE, assumption);
apply(assumption);
apply(drule greedy_not_exch);
by (erule notE, assumption);


theorem max_range_length : " \<lbrakk> max_range s; max_range t \<rbrakk> \<Longrightarrow> length s = length t "
apply(simp add: max_range_def);
apply(clarify);
apply(drule_tac x="t" in spec);
apply(drule_tac x="s" in spec);
by simp;



subsection {* Order properties of greedy *}


lemma V_ord_mono2  : " \<forall> a b c d. a << b \<and> c << d \<longrightarrow> (a +++ c) << (b +++ d) " 
apply(clarify);
apply(drule_tac d=c in V_ord_mono[rule_format]);
apply(drule_tac d=b in V_ord_mono[rule_format]);
apply(insert V2);
apply(frule_tac x=b in spec);
apply(drule_tac x=b in spec);
apply(drule_tac x=c in spec);
apply(drule_tac x=d in spec);
apply(erule ssubst);
back;
apply(drule subst, assumption);
apply(rule V_ord_trans[rule_format, OF conjI]);
apply(assumption);
back;
apply(assumption);
done



consts
Val :: "D list \<Rightarrow> V"

primrec
"Val [] = V_unit"
"Val (x#xs) = val x +++ Val xs" 

lemma Val_R : "Val (xs @ [x]) = Val xs +++ val x"
apply(induct_tac xs);
apply(simp add: V1 V2);
apply(simp add: V3);
done

lemma Val_app[rule_format] : " \<forall> ys. Val (xs @ ys) = Val xs +++ Val ys "
apply(induct_tac xs);
apply(simp add: V1 V2);
apply(simp add: V3);
done


lemma Val_subst[rule_format] : " y mem xs \<longrightarrow> Val (x#xs) = Val (y#(subst x y xs)) " 
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(case_tac "a = y");
apply(simp add: V3);
apply(insert V2);
apply(drule_tac x="val x" in spec);
apply(drule_tac x="val y" in spec);
apply(simp);
apply(drule mp, simp);
apply(simp);
apply(frule_tac x="val x" in spec);
apply(drule_tac x="val a" in spec);
apply(drule_tac x="Val list" in spec);
apply(drule_tac x="Val (subst x y list)" in spec);
apply(rule HOL.trans);
apply(rule V3[rule_format]);
apply(erule ssubst);
back;
apply(erule ssubst);
apply(erule ssubst);
apply(rule HOL.trans);
prefer 2; 
apply(rule V3[rule_format, THEN sym]);
apply(rule refl);
done

lemma Val_perm[rule_format] : " \<forall> ys. xs <~~> ys \<longrightarrow> Val xs = Val ys "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(case_tac ys);
apply(simp);
apply(clarify);
apply(case_tac "a = aa");
apply(clarify);
apply(simp);
apply(subgoal_tac "a mem lista");
apply(frule_tac x=aa in hd_swap[rule_format, OF conjI]);
apply(assumption);
apply(drule_tac x="subst aa a lista" in spec);
apply(drule mp);
apply(rule cons_perm_imp_perm);
apply(rule Permutation.trans);
apply(assumption);
apply(rule perm_sym);
apply(assumption);
apply(drule_tac x=aa in Val_subst);
apply(simp);
apply(subgoal_tac "a mem a#list");
apply(drule perm_mem, assumption);
apply(simp add: set_mem_eq);
by simp;

lemma greedy_Val2: "\<forall> xs P N N'. Val (fst (greedy P N xs)) = Val (fst (greedy P N' xs)) " 
apply(rule allI);
apply(induct_tac xs);
by simp_all;


lemma greedy_length1[rule_format, OF conjI] : 
                       " \<forall> P P' N. Feasible P \<and> P <~~> P' \<longrightarrow> 
                         length (fst (greedy P N xs)) = length (fst (greedy P' N xs)) "
apply(induct_tac xs);
apply(clarify);
apply(simp);
apply(erule perm_length);
apply(clarify);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(rule conjI);
apply(rule impI);
apply(drule_tac x="P @ [a]" in spec);
apply(drule_tac x="P' @ [a]" in spec);
apply(drule_tac x="N" in spec);
apply(drule mp);
apply(rule conjI);
apply(assumption);
apply(simp);
apply(assumption);
apply(rule impI);
apply(erule notE);
apply(erule FA2[rule_format, OF conjI]);
back;
apply(simp);
apply(rule impI);
apply(rule conjI);
apply(rule impI);
apply(erule notE);
apply(erule FA2[rule_format, OF conjI]);
back;
apply(rule perm_sym);
apply(simp);
apply(rule impI);
apply(drule_tac x="P" in spec);
apply(drule_tac x="P'" in spec);
apply(drule_tac x="a#N" in spec);
apply(drule mp, simp);
by assumption;

lemma greedy_length2: "\<forall> xs P N N'. length (fst (greedy P N xs)) = length (fst (greedy P N' xs)) " 
apply(rule allI);
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(drule_tac x="P @ [a]" in spec);
apply(drule_tac x="N" in spec);
apply(drule_tac x="N'" in spec);
apply(assumption);
apply(rule impI);
apply(drule_tac x="P" in spec);
apply(drule_tac x="a#N" in spec);
apply(drule_tac x="a#N'" in spec);
apply(assumption);
done


lemma Feas_perm_aux : " P @ [a, b] <~~> P @ [b, a] "
apply(simp);
apply(rule perm_append_Cons[of "a" "[b]" "[]", simplified]);
done


lemma Feas_last_swap : " Feasible (P @ [a, b]) \<Longrightarrow> Feasible (P @ [b, a]) "
apply(erule FA2[rule_format, OF conjI]);
apply(rule Feas_perm_aux);
done



lemma greedy_length3[rule_format] : " \<forall> P N. Feasible (P @ [x]) \<longrightarrow> 
                                        length P < length (fst (greedy (P @ [x]) N xs))"
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp);
apply(rule impI);
apply(drule_tac x="P @ [a]" in spec);
apply(drule mp);
apply(simp);
apply(erule Feas_last_swap);
apply(drule_tac x="N" in spec);
apply(simp);
apply(drule greedy_length1);
back;
apply(rule Feas_perm_aux);
apply(erule ssubst);
by simp;


theorem greedy_Val_ord : " \<forall> P P' N ys. Feasible P \<and> distinct (P @ xs)
                            \<and> sublist (P' @ ys) (P @ xs @ N) \<and> Feasible (P' @ ys)                  
                            \<and> length P = length P' \<and> Val P << Val P'
                            \<and> length (fst (greedy P N xs)) = length (P' @ ys)
                            \<and> sorted op <<* (P @ xs)
                            \<and> sorted op <<* (P' @ ys)
                            \<and> sorted op <<* (N @ xs)
                            \<and> (\<forall> n \<in> set N. (\<not> Feasible (P @ [n])))
                          \<longrightarrow> (Val (fst (greedy P N xs))) << (Val (P' @ ys)) "  
apply(induct_tac xs);
apply(clarify);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(simp only: greedy.simps);
apply(split split_if_asm);
apply(rule conjI);
apply(rule impI);
apply(case_tac ys);
apply(drule_tac P=P and x=a and N=N and xs=list in greedy_length3);
apply(simp);
apply(case_tac "val a << val aa");
apply(drule_tac x="P @ [a]" in spec);
apply(drule_tac x="P' @ [aa]" in spec);
apply(drule_tac x="N" in spec);
apply(drule_tac x="lista" in spec);
apply(drule mp);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(rule_tac t="(P @ [a]) @ list" and s="P @ a # list" in ssubst);
apply(simp (no_asm));
apply(assumption);
apply(rule conjI);
apply(simp (no_asm));
apply(clarify);
apply(rule_tac t="P @ a # list @ N" and s="P @ (a # list) @ N" in ssubst);
apply(simp (no_asm));
apply(assumption);
apply(rule conjI);
apply(simp (no_asm));
apply(clarify);
apply(rule conjI);
apply(simp (no_asm));
apply(rule conjI);
apply(simp (no_asm) add: Val_R);
apply(rule V_ord_mono2[rule_format]);
apply(rule conjI);
apply(assumption);
apply(assumption);
apply(rule conjI);
apply(simp (no_asm));
apply(clarify);
apply(rule_tac t="Suc (length P' + length lista)" and s="length (P' @ aa # lista)" in ssubst);
apply(simp (no_asm));
apply(assumption);
apply(rule conjI);
apply(rule_tac t="(P @ [a]) @ list" and s="P @ a # list" in ssubst);
apply(simp (no_asm));
apply(assumption);
apply(rule conjI);
apply(clarify);
apply(rule_tac t="(P' @ [aa]) @ lista" and s="P' @ aa # lista" in ssubst);
apply(simp (no_asm));
apply(assumption);
apply(rule conjI);
apply(simp only: sorted_append);
apply(simp only: sorted.simps);
apply(elim conjE);
apply(rule conjI);
apply(simp (no_asm));
apply(rule conjI);
apply(simp (no_asm));
apply(clarify);
apply(drule_tac x=x in bspec, assumption);
back;
apply(drule_tac x=y and A="set (a # list)" in bspec);
apply(simp (no_asm));
apply(rule disjI2);
apply(assumption);
apply(assumption);
apply(rule ballI);
apply(rule notI);
apply(drule_tac x=n in bspec, assumption);
apply(erule_tac P="Feasible (P @ [n])" in notE);
apply(rule_tac B="(P @ [a]) @ [n]" in M2[rule_format]);
apply(rule conjI);
apply(drule_tac A="(P @ [a]) @ [n]" in FA1[rule_format]);
apply(elim conjE);
apply(drule distinct_append[THEN iffD1]);
back;
apply(elim conjE);
apply(rule distinct_append[THEN iffD2]);
apply(rule conjI);
apply(drule distinct_append[THEN iffD1]);
back;
apply(elim conjE);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(simp only: Int_def);
apply(rule equalityI);
apply(rule subsetI);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(erule conjE);
apply(erule equalityE);
apply(drule_tac c=x in subsetD);
apply(simp (no_asm));
apply(rule conjI);
apply(rule disjI2);
apply(assumption);
apply(simp only: set.simps);
apply(drule singletonD);
apply(assumption);
apply(assumption);
apply(simp (no_asm));
apply(rule conjI);
apply(assumption);
apply(rule sublist_subset[THEN iffD2]);
apply(rule subsetI);
apply(simp only: set_append);
apply(simp (no_asm));
apply(simp only: Un_def);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(erule disjE);
apply(rule disjI2);
apply(rule disjI2);
apply(assumption);
apply(rule disjI1);
apply(simp only: set.simps);
apply(drule singletonD);
apply(assumption);
apply(clarify);
apply(rule_tac s="(P' @ [aa]) @ lista" and t="P' @ aa # lista" in ssubst);
apply(simp (no_asm));
apply(assumption);
apply(case_tac "val a = val aa");
apply(erule_tac P="val a << val aa" in notE);
apply(erule_tac t="val aa" in subst);
apply(simp (no_asm) add: V_ord_refl);
apply(case_tac "a = aa");
apply(erule_tac P="val a = val aa" in notE);
apply(rule_tac f=val in arg_cong);
apply(assumption);
apply(subgoal_tac "Feasible (P' @ [aa])");
prefer 2;
apply(rule_tac B="P' @ ys" in M2[rule_format]);
apply(rule conjI);
apply(drule_tac A="P' @ ys" in FA1[rule_format]);
apply(clarify);
apply(drule distinct_append[THEN iffD1]);
back;
apply(elim conjE);
apply(rule distinct_append[THEN iffD2]);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(simp (no_asm));
apply(simp only: Int_def);
apply(rule equalityI);
apply(rule subsetI);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(erule conjE);
apply(erule equalityE);
apply(drule_tac c=x in subsetD);
apply(simp (no_asm));
apply(rule conjI);
apply(assumption);
apply(rule disjI1);
apply(simp only: set.simps);
apply(drule singletonD);
apply(assumption);
apply(assumption);
apply(simp (no_asm));
apply(rule conjI);
apply(assumption);
apply(rule sublist_subset[THEN iffD2]);
apply(rule subsetI);
apply(simp only: set_append);
apply(simp only: Un_def);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(rule mem_Collect_eq[THEN iffD2]);
apply(erule disjE);
apply(rule disjI1);
apply(assumption);
apply(rule disjI2);
apply(simp (no_asm));
apply(rule disjI1);
apply(simp only: set.simps);
apply(drule singletonD);
apply(assumption);
apply(drule_tac A=P and B="P' @ [aa]" in M3[rule_format, OF conjI]);
apply(rule conjI);
apply(assumption);
apply(erule_tac t="length P" in ssubst);
apply(simp (no_asm));
apply(erule exE);
apply(elim conjE);
apply(subgoal_tac "e \<notin> set N");
prefer 2;
apply(rule notI);
apply(drule_tac x=e in bspec, assumption);
apply(clarify);
apply(drule_tac x1=e in set_mem_eq[THEN iffD1]);
apply(subgoal_tac "e \<in> set (a#list)");
prefer 2;
apply(drule sublist_subset[THEN iffD1]);
apply(drule_tac c=e in subsetD);
apply(clarify);
apply(simp only: set_append);
apply(simp only: Un_def);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(rule mem_Collect_eq[THEN iffD2]);
apply(erule disjE);
apply(rule disjI1);
apply(assumption);
apply(rule disjI2);
apply(simp only: set.simps);
apply(drule singletonD);
apply(erule_tac t=e in ssubst);
apply(simp (no_asm));
apply(simp only: set_append);
apply(simp only: Un_def);
apply(drule mem_Collect_eq[THEN iffD1]);
back;
apply(erule disjE);
apply(erule_tac P="e mem P" in notE);
apply(rule set_mem_eq[THEN iffD2]);
apply(assumption);
apply(drule mem_Collect_eq[THEN iffD1]);
back;
apply(erule disjE);
apply(assumption);
apply(clarify);
apply(simp only: set_append);
apply(simp only: set.simps);
apply(drule insert_iff[THEN iffD1]);
apply(simp only: Un_def);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(erule disjE);
apply(erule disjE);
apply(drule sorted_append[THEN iffD1]);
back;
apply(elim conjE);
apply(drule_tac x=e in bspec, assumption);
apply(drule_tac x=aa in bspec);
back;
apply(simp (no_asm));
apply(simp only: D_ord_def);
apply(drule singletonD);
apply(clarify);
apply(erule disjE);
apply(drule sorted_append[THEN iffD1]);
apply(drule sorted_append[THEN iffD1]);
apply(elim conjE);
apply(drule_tac x=e in bspec, assumption);
apply(drule_tac x=aa in bspec);
back;
back;
apply(simp (no_asm));
apply(drule sorted.simps[THEN iffD1]);
apply(erule conjE);
apply(drule_tac x=e in bspec, assumption);
apply(simp only: D_ord_def);
apply(erule_tac P="val a << val aa" in notE);
apply(rule V_ord_trans[rule_format]);
apply(rule conjI);
apply(assumption);
apply(assumption);
apply(drule singletonD);
apply(drule sorted_append[THEN iffD1]);
apply(elim conjE);
apply(drule sorted.simps[THEN iffD1]);
apply(elim conjE);
apply(drule_tac x=e in bspec, assumption);
apply(erule_tac P="val a << val aa" in notE);
apply(simp only: D_ord_def);
apply(rule impI);
apply(clarify);
apply(rule conjI);
apply(rule impI);
apply(clarify);
apply(rule impI);
apply(drule_tac x="P" in spec);
apply(drule_tac x="P'" in spec);
apply(drule_tac x="N @ [a]" in spec);
apply(drule_tac x="ys" in spec);
apply(drule mp);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(drule distinct_append[THEN iffD1]);
apply(rule distinct_append[THEN iffD2]);
apply(elim conjE);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(simp only: distinct.simps);
apply(rule equalityI);
apply(rule subsetI);
apply(simp only: Int_def);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(erule conjE);
apply(erule equalityE);
apply(drule_tac c=x in subsetD);
apply(rule mem_Collect_eq[THEN iffD2]);
apply(rule conjI);
apply(assumption);
apply(simp (no_asm));
apply(rule disjI2);
apply(assumption);
apply(assumption);
apply(simp (no_asm));
apply(rule conjI);
apply(rule sublist_subset[THEN iffD2]);
apply(drule sublist_subset[THEN iffD1]);
apply(rule subsetI);
apply(drule_tac c=x in subsetD, assumption);
apply(simp only: set_append);
apply(simp only: Un_def);
apply(drule mem_Collect_eq[THEN iffD1]);
back;
apply(rule mem_Collect_eq[THEN iffD2]);
apply(erule disjE);
apply(rule disjI1);
apply(assumption);
apply(drule mem_Collect_eq[THEN iffD1]);
back;
apply(rule disjI2);
apply(rule mem_Collect_eq[THEN iffD2]);
apply(erule disjE);
apply(simp only: set.simps);
apply(drule insert_iff[THEN iffD1]);
apply(erule disjE);
apply(rule disjI2);
apply(rule mem_Collect_eq[THEN iffD2]);
apply(rule disjI2);
apply(erule_tac t=x in ssubst);
apply(simp (no_asm));
apply(rule disjI1);
apply(assumption);
apply(rule disjI2);
apply(rule mem_Collect_eq[THEN iffD2]);
apply(rule disjI1);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(insert greedy_length2);
apply(drule_tac x=list in spec);
apply(drule_tac x=P in spec);
apply(drule_tac x="a#N" in spec);
apply(drule_tac x="N@[a]" in spec);
apply(simp del: sorted_append distinct_append);
apply(rule conjI);
apply(rule sorted_append[THEN iffD2]);
apply(drule sorted_append[THEN iffD1]);
apply(elim conjE);
apply(drule sorted.simps[THEN iffD1]);
apply(elim conjE);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(clarify);
apply(drule_tac x=x in bspec, assumption);
apply(drule_tac x=y in bspec);
back;
back;
apply(simp (no_asm));
apply(rule disjI2);
apply(assumption);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(rule_tac t="(N @ [a]) @ list" and s="N @ a # list" in ssubst);
apply(simp (no_asm));
apply(assumption);
apply(rule ballI);
apply(case_tac "n = a");
apply(clarify);
apply(drule_tac x=n in bspec);
apply(simp only: set_append);
apply(simp only: Un_def);
apply(drule mem_Collect_eq[THEN iffD1]);
apply(erule disjE);
apply(assumption);
apply(simp only: set.simps);
apply(drule singletonD);
apply(erule_tac P="n = a" in notE);
apply(assumption);
apply(assumption);
apply(insert greedy_Val2);
apply(drule_tac x=list in spec);
back;
apply(drule_tac x=P in spec);
back;
apply(drule_tac x="a#N" in spec);
back;
apply(drule_tac x="N @ [a]" in spec);
back;
apply(erule ssubst, assumption);
done


section {* Main result *}

consts
canonical_greedy :: "D list \<Rightarrow> D list \<Rightarrow> D list"
primrec
"canonical_greedy P [] = P"
"canonical_greedy P (x#xs) = (if (Feasible (P @ [x])) then canonical_greedy (P @ [x]) xs
                                                      else canonical_greedy P xs)"

lemma c_greedy_eq[rule_format] : " \<forall> P N. canonical_greedy P xs = fst (greedy P N xs) "
apply(induct_tac xs);
by simp_all;


theorem greedy_main_prop : " max_range (canonical_greedy [] U) \<and>
                             (\<forall> s. max_range s \<longrightarrow> Val (canonical_greedy [] U) << Val s) "
apply(rule_tac t="canonical_greedy [] U" in subst);
apply(rule sym);
apply(rule_tac N="[]" in c_greedy_eq);
apply(insert greedy_max_range);
apply(rule conjI);
apply(assumption);
apply(clarify);
apply(frule max_range_length);
apply(assumption);
apply(insert A_sort);
apply(drule_tac x=s in spec);
apply(erule conjE);
apply(frule Val_perm);
apply(erule_tac t="Val s" in ssubst);
apply(rule greedy_Val_ord[rule_format, of _ _ "[]", simplified]);
apply(rule conjI);
apply(rule M1);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(rule U1);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(simp (no_asm) add: sublist_def);
apply(clarify);
apply(drule perm_sym);
apply(unfold max_range_def);
apply(elim conjE);
apply(drule_tac A=s in FA1[rule_format]);
apply(erule conjE);
apply(drule perm_mem, assumption);
apply(drule sublistD, assumption);
apply(assumption);
apply(rule conjI);
apply(elim conjE);
apply(rule FA2[rule_format]);
apply(rule conjI);
prefer 2;
apply(assumption);
apply(assumption);
apply(fold max_range_def);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(simp add: V_ord_refl);
apply(rule conjI);
apply(drule max_range_length, assumption);
apply(drule perm_length);
apply(simp);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(rule U2);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(rule U2);
apply(rule conjI);
apply(simp);
apply(simp);
done


end