
theory Greedy_Aux2 = Greedy_Aux : 

subsection {* sublist and rem properties *}

constdefs
count :: "'a \<Rightarrow> 'a list \<Rightarrow> nat"
"count x l \<equiv> length (pos 0 x l)"

consts
rem_all :: "'a \<Rightarrow> 'a list \<Rightarrow> 'a list "
primrec
"rem_all x [] = []"
"rem_all x (y#ys) = (if (x = y) then (rem_all x ys) else (y#(rem_all x ys)))"

constdefs
f_rem :: " 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list "
"f_rem xs l \<equiv> foldr rem xs l"

f_rem2 :: " 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list "
"f_rem2 xs l \<equiv> foldl (\<lambda> l x. rem x l) l xs"


theorem rem_swap : " rem x (rem y xs) = rem y (rem x xs) "
apply(induct_tac xs);
by simp_all;

lemma f_rem2_aux : "\<forall> ys. foldr rem xs (rem a ys) = rem a (foldr rem xs ys)"
apply(induct_tac xs);
apply(simp);
apply(rule allI);
apply(simp);
by (rule rem_swap);

lemma f_rem2_eq : " \<forall> xs l. f_rem2 xs l = f_rem xs l "
apply(rule allI);
apply(induct_tac xs);
apply(simp add: f_rem2_def f_rem_def);
apply(rule allI);
apply(simp add: f_rem2_def f_rem_def);
apply(rule f_rem2_aux[rule_format]);
done


consts
duplicate :: " 'a \<Rightarrow> nat \<Rightarrow> 'a list "
primrec
"duplicate x 0 = []"
"duplicate x (Suc n) = x#(duplicate x n)"

 

consts
rem_n :: "nat \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list "
primrec 
"rem_n n x [] = []"
"rem_n n x (y#ys) = (case n of
                      0 \<Rightarrow> (y#ys)
                    | Suc n' \<Rightarrow> (if (x = y) then (rem_n n' x ys) else (y#(rem_n n x ys)))) " 


constdefs
rem_count :: "'a \<Rightarrow> (nat \<times> 'a list) \<Rightarrow> (nat \<times> 'a list)"
"rem_count x \<equiv> \<lambda> (n, l). (n + count x l, rem_all x l)"

f_rem_count :: "'a list \<Rightarrow> (nat \<times> 'a list)"
"f_rem_count l \<equiv> foldr rem_count l (0, l) "


lemma rem_all1[rule_format] : " (\<not> x mem l) \<longrightarrow> rem_all x l = l "
apply(induct_tac l);
by simp_all;

lemma rem_all2[rule_format] : "\<forall> x.  x mem l \<longrightarrow> (\<not>(x mem (rem_all x l))) " 
apply(induct_tac l);
apply(simp);
apply(clarify);
apply(case_tac "x = a");
apply(clarify);
apply(case_tac "a mem list");
apply(drule_tac x=a in spec);
apply(drule mp, simp);
apply(simp);
apply(drule rem_all1);
apply(drule_tac x=a in spec);
apply(drule mp, simp);
apply(simp);
apply(drule_tac x=x in spec);
apply(drule mp, simp add: set_mem_eq);
by (simp add: set_mem_eq);

lemma rem_all3[rule_format] : "\<forall> x.  x mem l \<longrightarrow> length (rem_all x l) < length l " 
apply(induct_tac l);
apply(simp);
apply(clarify);
apply(case_tac "x = a");
apply(case_tac "a mem list");
apply(drule_tac x=a in spec);
apply(drule mp, simp);
apply(simp);
apply(drule rem_all1);
apply(simp);
apply(drule_tac x=x in spec);
apply(drule mp, simp add: set_mem_eq);
by simp;


lemma rem_all4[simp] : "rem_all x (rem_all x l) = rem_all x l"
apply(case_tac "x mem l");
apply(drule rem_all2);
apply(drule rem_all1);
apply(assumption);
apply(drule rem_all1);
by simp;

lemma rem_all_count[simp] : " count x (rem_all x l) = 0 "
apply(case_tac "x mem l");
apply(drule rem_all2);
apply(simp add: count_def);
apply(drule pos2_cps);
apply(assumption);
apply(frule rem_all1);
apply(erule ssubst);
apply(drule pos2_cps);
by (simp add: count_def);


lemma count_rem[rule_format] : " x mem l \<longrightarrow> count x l = Suc(count x (rem x l)) "
apply(induct_tac l);
apply(simp);
apply(rule impI);
apply(case_tac "x = a");
apply(clarify);
apply(simp add: count_def);
apply(simp add: pos_r_shift[rule_format, OF refl]);
apply(drule mp, simp add: set_mem_eq);
apply(simp add: count_def);
by (simp add: pos_r_shift[rule_format, OF refl]);


lemma rem_n_0[simp] : "rem_n 0 x l = l"
apply(induct_tac l);
by simp_all;

lemma rem_n_1 : "rem_n 1 x l = rem x l"
apply(induct_tac l);
by simp_all;

lemma rem_n_count : "rem_n (count x l) x l = rem_all x l"
apply(induct_tac l);
apply(simp add: count_def);
apply(simp add: count_def);
apply(rule conjI);
apply(clarify);
apply(simp add: pos_r_shift[rule_format, OF refl]);
apply(rule impI);
apply(split nat.split);
apply(rule conjI);
apply(rule impI);
apply(simp (no_asm));
apply(erule subst);
apply(simp add: pos_r_shift[rule_format, OF refl]);
apply(clarify);
apply(split split_if);
apply(rule conjI);
apply(simp);
apply(rule impI);
apply(simp (no_asm));
apply(erule subst);
apply(simp add: pos_r_shift[rule_format, OF refl]);
done

lemma rem_rem_n[rule_format] : "\<forall> n. rem x (rem_n n x xs) = rem_n (Suc n) x xs "
apply(induct_tac xs);
apply(simp);
apply(rule allI);
apply(simp);
apply(rule conjI);
apply(rule impI);
apply(split nat.split);
apply(rule conjI);
apply(simp);
apply(clarify);
apply(split split_if);
apply(rule conjI);
apply(simp);
apply(clarify);
apply(clarify);
apply(split nat.split);
apply(rule conjI);
apply(clarify);
apply(simp);
apply(drule_tac x=0 in spec);
apply(simp);
apply(clarify);
apply(split split_if);
apply(rule conjI);
apply(clarify);
apply(clarify);
by simp;


lemma f_rem_rem_n : " \<forall> l. f_rem (duplicate x n) l = rem_n n x l "
apply(induct_tac n);
apply(simp add: f_rem_def);
apply(rule allI);
apply(simp add: f_rem_def);
by (rule rem_rem_n);

lemma f_rem_rem_all : " f_rem (duplicate x (count x l)) l = rem_all x l "
apply(rule HOL.trans);
apply(rule f_rem_rem_n[rule_format]);
by (rule rem_n_count);



(*
lemma length_rem_count : " length l = count x l + (length (rem_all x l)) "
*)




lemma f_rem_all[simp] : " f_rem xs xs = [] "
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(insert f_rem2_eq);
apply(frule_tac x="a#list" in spec);
apply(drule_tac x="a#list" in spec);
back;
apply(erule subst);
back;
back;
apply(simp (no_asm) add: f_rem2_def);
apply(fold f_rem2_def);
apply(drule_tac x=list in spec);
apply(drule_tac x=list in spec);
by simp;


theorem f_rem_sorted : " sorted le l \<Longrightarrow> sorted le (f_rem xs l) "
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(simp add: f_rem_def);
by (erule rem_sorted);


theorem f_rem_sublist : " \<forall> l xs. sublist (f_rem xs l) l "
apply(intro allI);
apply(induct_tac xs);
apply(simp add: f_rem_def sublist_def);
apply(simp add: f_rem_def);
apply(rule sublist_trans);
apply(rule rem_sub);
by assumption;

lemma f_rem_perm_aux[rule_format, OF conjI] : 
                        "x \<noteq> y \<and> x mem xs 
                         \<longrightarrow> rem x (foldr rem (subst y x xs) l) = rem y (foldr rem xs l) "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(rule rem_swap);
apply(rule impI);
apply(drule mp, simp);
apply(rule HOL.trans);
apply(rule rem_swap);
apply(rule HOL.trans);
prefer 2;
apply(rule rem_swap);
by simp;


theorem f_rem_perm : " \<forall> ys. xs <~~> ys \<longrightarrow> f_rem xs l = f_rem ys l "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(case_tac ys);
apply(simp);
apply(clarify);
apply(case_tac "a = aa");
apply(clarify);
apply(simp add: f_rem_def);
apply(simp add: f_rem_def);
apply(subgoal_tac "a mem lista");
apply(frule hd_swap[rule_format, OF conjI]);
apply(assumption);
apply(drule_tac x="subst aa a lista" in spec);
apply(drule mp);
apply(drule perm_sym);
back;
apply(drule Permutation.trans, assumption);
apply(simp);
apply(drule f_rem_perm_aux);
apply(assumption);
apply(erule ssubst);
apply(assumption);
apply(drule_tac x=a in perm_mem);
apply(simp);
apply(simp add: set_mem_eq);
done



theorem f_rem_perm2[rule_format] : "l <~~> r \<longrightarrow> f_rem xs l <~~> f_rem xs r "
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(clarify);
apply(simp (no_asm) add: f_rem_def);
apply(fold f_rem_def);
apply(rule rem_perm2);
by assumption;


lemma f_rem_dist : " distinct l \<Longrightarrow> distinct (f_rem xs l) "
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(simp add: f_rem_def);
apply(case_tac "a mem (foldr rem list l)");
apply(erule rem_dist3, assumption);
apply(drule rem_dist1);
by simp;

lemma f_rem_mem[rule_format] : " x mem (f_rem xs l) \<longrightarrow> x mem l " 
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(rule impI);
apply(simp add: f_rem_def);
apply(drule rem_aux);
by simp;


lemma f_rem_perm3[rule_format, OF conjI] : 
             " \<forall> l. distinct xs \<and> sublist xs l \<longrightarrow> l <~~> xs @ (f_rem xs l) "
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(clarify);
apply(simp add: f_rem_def);
apply(simp add: f_rem2_aux[rule_format, THEN sym]);
apply(subgoal_tac "a mem l");
apply(drule rem_perm);
apply(rule Permutation.trans);
apply(rule perm_sym);
apply(assumption);
apply(drule_tac x="rem a l" in spec);
apply(drule mp);
apply(erule conjE);
apply(simp add: sublist_def);
apply(clarify);
apply(case_tac "a = x");
apply(simp add: set_mem_eq);
apply(drule_tac x=x in spec);
apply(drule mp);
apply(split split_if);
apply(simp);
apply(erule rem_aux2, assumption);
apply(simp);
apply(simp add: sublist_def);
done


lemma f_rem_length : 
             " \<lbrakk> distinct xs; sublist xs l \<rbrakk> 
                  \<Longrightarrow> length l = length xs + length (f_rem xs l) "
apply(drule f_rem_perm3);
apply(assumption);
apply(drule perm_length);
by simp;

lemma f_rem_length2 : 
             " \<lbrakk> distinct xs; sublist xs l \<rbrakk> 
                  \<Longrightarrow> (length xs \<le> length l) \<and> ((length (f_rem xs l)) \<le> length l) \<and>
                                    (length (f_rem xs l) = length l - length xs) "
by(simp add: f_rem_length);




constdefs
rem_sublist :: " 'a list \<Rightarrow> 'a list \<Rightarrow> bool "
"rem_sublist l r \<equiv> (\<exists> xs. sublist xs r \<and> l <~~> f_rem xs r)"


theorem rem_sublist_is_sublist : " rem_sublist l r \<Longrightarrow> sublist l r "
apply(simp add: rem_sublist_def);
apply(clarify);
apply(rule sublist_perm);
apply(rule f_rem_sublist[rule_format]);
apply(rule perm_sym);
by assumption;


consts
ldiff :: " 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list "
primrec
"ldiff [] ys = []"
"ldiff (x#xs) ys = (if (x mem ys) then (ldiff xs ys) else (x#(ldiff xs ys)))"

lemma ldiff0[simp] : " ldiff xs [] = xs "
apply(induct_tac xs);
by simp_all;


lemma ldiff1a[rule_format] : " x mem ldiff xs ys \<longrightarrow> (x mem xs \<and> (\<not> x mem ys)) "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp only: ldiff.simps);
apply(split split_if_asm);
apply(drule mp, simp);
apply(simp);
apply(case_tac "a = x");
apply(simp);
apply(drule mp, simp);
by simp;

lemma ldiff1a_cps1 : " x \<notin> set xs \<Longrightarrow> (\<not> (x mem ldiff xs ys)) "
apply(rule notI);
apply(drule ldiff1a);
by (simp add: set_mem_eq);


lemma ldiff1a_cps2 : " x mem ys \<Longrightarrow> (\<not> (x mem ldiff xs ys)) "
apply(rule notI);
apply(drule ldiff1a);
by (simp);




lemma ldiff1b[rule_format, OF conjI] : " (x mem xs \<and> (\<not> x mem ys)) \<longrightarrow> x mem ldiff xs ys "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(case_tac "a = x");
apply(clarify);
by simp_all;


lemma ldiff_sublist1[rule_format] : "\<forall> ys. ldiff xs ys = [] \<longrightarrow> sublist xs ys "
apply(induct_tac xs);
apply(simp add: sublist_def);
apply(clarify);
apply(simp);
apply(split split_if_asm);
apply(drule_tac x=ys in spec);
apply(drule mp, simp);
apply(rule sublist_subset[THEN iffD2]);
apply(drule sublist_subset[THEN iffD1]);
apply(rule subsetI);
apply(simp);
apply(erule disjE);
apply(simp add: set_mem_eq);
apply(erule subsetD, assumption);
by simp;

lemma ldiff_sublist2[rule_format] : "\<forall> ys. sublist xs ys \<longrightarrow> ldiff xs ys = [] "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp);
apply(rule conjI);
apply(rule impI);
apply(drule_tac x=ys in spec);
apply(drule mp);
apply(simp add: sublist_def);
apply(assumption);
apply(erule sublistD);
by simp;

lemma ldiff_sublist3 : " \<forall> xs ys. sublist (ldiff xs ys) xs " 
apply(clarify);
apply(simp add: sublist_def);
apply(clarify);
apply(drule ldiff1a);
by simp;



lemma ldiff2 : " \<not> x mem xs \<Longrightarrow> (\<not> (x mem ldiff xs ys))" 
apply(rule notI);
apply(drule ldiff1a);
by (simp);



lemma ldiff_red[rule_format] : " y \<notin> set xs \<longrightarrow> ldiff xs (y#ys) = ldiff xs ys" 
apply(induct_tac xs);
apply(simp);
apply(rule impI);
apply(drule mp, simp);
by simp;

consts
ldiff2 :: " 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list "
primrec
"ldiff2 [] ys = []"
"ldiff2 (x#xs) ys = (if (x mem ys) then (ldiff2 xs (rem x ys)) else (x#(ldiff2 xs ys)))"


lemma ldiff_red2[rule_format, OF conjI] : " \<forall> ys. a \<notin> set xs \<and> a mem ys \<and> distinct xs
                                            \<longrightarrow> ldiff xs ys = ldiff xs (rem a ys) "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(rule conjI);
apply(rule impI);
apply(drule_tac x=ys in spec);
apply(drule mp, simp);
apply(assumption);
apply(rule impI);
apply(simp);
apply(clarify);
apply(drule rem_aux2, assumption);
apply(simp);
apply(rule impI);
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(clarify);
apply(erule notE);
apply(rule_tac x=a in rem_aux_spec);
apply(assumption);
apply(assumption);
apply(rule impI);
apply(drule_tac x=ys in spec);
apply(drule mp, simp);
apply(assumption);
done



lemma ldiff2_ldiff : " \<forall> xs ys. distinct xs \<longrightarrow> ldiff2 xs ys = ldiff xs ys "
apply(rule allI);
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp);
apply(rule impI);
apply(rule sym);
apply(rule ldiff_red2);
by simp_all;





lemma ldiff_subst[rule_format] : " x mem ys \<longrightarrow> ldiff (y#ys) l <~~> ldiff (x#(subst y x ys)) l " 
apply(induct_tac ys);
apply(simp);
apply(rule impI);
apply(case_tac "a = x");
apply(clarify);
apply(simp (no_asm));
apply(clarify);
apply(drule mp, simp);
apply(simp (no_asm));
apply(rule conjI);
apply(clarify);
apply(clarify);
apply(rule conjI);
apply(clarify);
apply(rule conjI);
apply(clarify);
apply(rule conjI);
apply(clarify);
apply(simp);
apply(clarify);
apply(simp);
apply(clarify);
apply(rule conjI);
apply(clarify);
apply(simp);
apply(clarify);
apply(simp);
apply(clarify);
apply(rule conjI);
apply(clarify);
apply(rule conjI);
apply(clarify);
apply(simp);
apply(clarify);
apply(simp);
apply(rule Permutation.trans);
prefer 2;
apply(rule Permutation.swap);
apply(simp);
apply(clarify);
apply(rule conjI);
apply(clarify);
apply(simp);
apply(rule Permutation.trans);
apply(rule Permutation.swap);
apply(simp);
apply(clarify);
apply(simp);
apply(rule Permutation.trans);
apply(rule Permutation.swap);
apply(rule Permutation.trans);
prefer 2;
apply(rule Permutation.swap);
apply(simp);
done


lemma ldiff_perm[rule_format] : " \<forall> ys. xs <~~> ys \<longrightarrow> ldiff xs l <~~> ldiff ys l "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(case_tac ys);
apply(simp);
apply(clarify);
apply(case_tac "a = aa");
apply(simp);
apply(subgoal_tac "a mem lista");
apply(frule hd_swap[rule_format, OF conjI]);
apply(assumption);
apply(drule_tac x="subst aa a lista" in spec);
apply(drule mp);
apply(rule cons_perm_eq[THEN iffD1]);
apply(rule Permutation.trans);
apply(assumption);
apply(rule perm_sym);
apply(assumption);
apply(drule_tac x=a and y=aa and l=l in ldiff_subst);
apply(rule Permutation.trans);
prefer 2;
apply(rule perm_sym);
apply(assumption);
apply(simp);
apply(subgoal_tac "a mem (a#list)");
apply(drule perm_mem, assumption);
apply(simp add: set_mem_eq);
by simp;

lemma ldiff_dist[rule_format] : " \<forall> ys. distinct xs \<longrightarrow> distinct (ldiff xs ys) "
apply(induct_tac xs);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(rule impI);
apply(rule conjI);
apply(simp);
apply(erule conjE);
apply(drule_tac ys=ys in ldiff1a_cps1);
apply(simp add: set_mem_eq);
by simp;




lemma ldiff2_perm : " \<lbrakk> distinct xs; xs <~~> ys \<rbrakk> \<Longrightarrow> ldiff2 xs l <~~> ldiff2 ys l"
apply(frule perm_dist, assumption);
apply(drule_tac ys=l in ldiff2_ldiff[rule_format]);
apply(drule_tac ys=l in ldiff2_ldiff[rule_format]);
apply(erule ssubst);
apply(erule ssubst);
by(erule ldiff_perm);

lemma ldiff21a_cps1 : " \<lbrakk> x \<notin> set xs; distinct xs \<rbrakk> 
                        \<Longrightarrow> (\<not> (x mem ldiff2 xs ys)) "
apply(drule ldiff2_ldiff[rule_format]);
apply(erule ssubst);
apply(erule ldiff1a_cps1);
done

lemma ldiff2_sublist3 : " distinct xs \<Longrightarrow> sublist (ldiff2 xs ys) xs " 
apply(drule ldiff2_ldiff[rule_format]);
apply(erule ssubst);
by (rule ldiff_sublist3[rule_format]);

lemma ldiff2_dist : " distinct xs \<Longrightarrow> distinct (ldiff2 xs ys) "
apply(frule ldiff_dist);
apply(drule ldiff2_ldiff[rule_format]);
apply(erule ssubst);
by assumption;



lemma f_rem_red[rule_format] : " \<not> y mem xs \<longrightarrow> f_rem xs (y#ys) = y#(f_rem xs ys) "
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(rule impI);
apply(case_tac "y = a");
apply(clarify);
apply(simp);
apply(simp add: f_rem_def);
apply(split split_if_asm);
apply(simp);
apply(drule mp, simp);
apply(erule ssubst);
by simp;

lemma f_rem_red2[rule_format, OF conjI] : " \<not> a mem xs \<and> a mem ys \<longrightarrow> a mem (f_rem xs ys) "
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(clarify);
apply(simp (no_asm) add: f_rem_def);
apply(fold f_rem_def);
apply(simp);
apply(split split_if_asm);
apply(simp);
apply(drule mp, simp);
apply(drule_tac x=aa in rem_aux2);
apply(assumption);
back;
by assumption;

lemma ldiff4[rule_format] : " \<forall> ys. distinct xs \<longrightarrow> ldiff (f_rem (ldiff xs ys) xs) ys = [] " 
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(rule allI);
apply(simp);
apply(rule conjI);
prefer 2;
apply(clarify);
apply(insert f_rem2_eq);
apply(drule_tac x="a # ldiff list ys" in spec);
back;
apply(drule_tac x="a # list" in spec);
back;
apply(erule subst);
apply(simp add: f_rem2_def);
apply(fold f_rem2_def);
apply(drule_tac x=ys in spec);
apply(simp add: f_rem2_eq);
apply(clarify);
apply(subgoal_tac "\<not> a mem list");
apply(drule_tac ys=ys in ldiff2);
apply(drule f_rem_red);
apply(erule ssubst);
apply(drule_tac x=ys in spec);
back;
apply(simp);
by (simp add: set_mem_eq);




theorem f_rem_perm_dist : "\<forall> l. distinct l \<and> distinct r \<and> sublist l r 
                           \<longrightarrow> l <~~> f_rem (ldiff2 r l) r"
apply(induct_tac r);
apply(clarify);
apply(simp add: sublist_def);
apply(case_tac l);
apply(simp);
apply(drule_tac x=a in spec);
apply(simp);
apply(clarify);
apply(case_tac l);
apply(simp);
apply(clarify);
apply(drule_tac ys="[]" in ldiff2_ldiff[rule_format]);
apply(simp);
apply(case_tac "a = aa");
apply(clarify);
apply(simp);
apply(clarify);
apply(frule_tac xs=list and ys=lista in ldiff2_ldiff[rule_format]); 
apply(erule ssubst);
apply(frule_tac ys=lista in ldiff1a_cps1);
back;
apply(drule_tac ys=list in f_rem_red);
apply(erule ssubst);
apply(frule_tac xs=list and ys=lista in ldiff2_ldiff[rule_format]); 
apply(erule subst);
apply(drule_tac x=lista in spec);
apply(drule mp);
apply(rule conjI);
apply(assumption);
apply(rule sublist_subset[THEN iffD2]);
apply(drule sublist_subset[THEN iffD1]);
apply(rule subsetI);
apply(simp);
apply(drule subsetD, assumption);
apply(simp);
apply(erule disjE);
apply(clarify);
apply(assumption);
apply(simp);
apply(simp);
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(rule impI);
apply(rule conjI);
apply(rule impI);
apply(clarify);
apply(frule_tac xs=list and ys="(aa # rem a lista)" in ldiff2_ldiff[rule_format]); 
apply(erule ssubst);
apply(subgoal_tac "aa mem list");
apply(frule_tac x=aa in rem_perm);
prefer 2;
apply(subgoal_tac "aa mem (aa#lista)");
apply(drule sublistD, assumption);
apply(simp);
apply(simp (no_asm));
apply(drule perm_sym);
apply(drule_tac l="aa # (rem a lista)" in ldiff_perm);
apply(drule f_rem_perm[rule_format]);
apply(erule ssubst);
apply(subgoal_tac "distinct (aa # rem aa list)");
prefer 2;
apply(simp);
apply(rule conjI);
apply(drule_tac l=list and x=aa in rem_dist2, assumption);
apply(simp add: set_mem_eq);
apply(drule_tac l=list and x=aa in rem_dist3, assumption);
apply(assumption);
apply(frule_tac xs="aa # (rem aa list)" and ys="aa # (rem a lista)" in ldiff2_ldiff[rule_format]); 
apply(erule subst);
apply(frule_tac x=aa in rem_perm);
apply(drule_tac z1=a in cons_perm_eq[THEN iffD2]);
apply(drule perm_sym);
apply(drule_tac zs="aa # a # rem aa list" in Permutation.trans);
apply(rule Permutation.trans);
apply(rule Permutation.swap);
apply(simp);
apply(drule perm_sym);
apply(drule f_rem_perm2);
apply(rule Permutation.trans);
prefer 2;
apply(assumption);
apply(frule_tac a=a and x=aa in rem_aux3);
apply(simp (no_asm));
apply(simp add: set_mem_eq);
apply(erule conjE);
apply(frule_tac x=a and xs="rem aa list" and ys="rem a lista" in ldiff1a_cps1);
apply(frule_tac x=aa and xs="rem aa list" and ys="rem a lista" in ldiff1a_cps1);
apply(frule_tac xs="rem aa list" and ys="rem a lista" in ldiff2_ldiff[rule_format]);
apply(erule ssubst);
apply(drule_tac y=aa in f_rem_red);
apply(erule ssubst);
apply(simp (no_asm));
apply(drule_tac y=a in f_rem_red);
apply(erule ssubst);
apply(drule_tac x="aa#(rem a lista)" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(drule_tac a=aa and x=a in rem_aux3);
apply(simp add: set_mem_eq);
apply(rule rem_dist3);
apply(assumption);
apply(simp add: set_mem_eq);
apply(rule sublist_subset[THEN iffD2]);
apply(drule sublist_subset[THEN iffD1]);
apply(rule subsetI);
apply(simp);
apply(erule disjE);
apply(simp);
apply(drule_tac x1=x in set_mem_eq[THEN iffD2]);
apply(drule_tac l=lista in rem_dist4, assumption);
apply(erule conjE);
apply(simp add: set_mem_eq);
apply(drule subsetD, assumption);
back;
apply(simp);
apply(drule set_mem_eq[THEN iffD2]);
apply(drule set_mem_eq[THEN iffD2]);
apply(drule rem_perm);
apply(drule rem_perm);
apply(rule Permutation.trans);
apply(rule perm_sym);
apply(assumption);
apply(simp (no_asm));
apply(drule_tac ys=list in perm_sym);
apply(frule_tac xs=list and l="aa # rem a lista" in ldiff2_perm, assumption);
apply(drule_tac xs="ldiff2 ?F ?G" and l=list in f_rem_perm[rule_format]);
apply(drule_tac P="\<lambda> x. (aa # rem a lista) <~~> x" in subst, assumption);
apply(simp);
apply(frule_tac xs="rem aa list" in ldiff2_ldiff[rule_format]);
apply(drule_tac P="\<lambda> x. (aa # rem a lista) <~~> f_rem x list" in subst, assumption);
apply(subgoal_tac "aa mem list");
apply(drule_tac x=aa in rem_perm);
apply(drule_tac ys=list in perm_sym);
apply(drule_tac l=list in f_rem_perm2);
apply(drule_tac xs="aa # rem a lista" and ys="f_rem (ldiff ?F ?G) list" in Permutation.trans, assumption);
apply(drule_tac x=aa in ldiff1a_cps1);
back;
apply(drule_tac y=aa in f_rem_red);
apply(drule_tac P="\<lambda> x. (aa # rem a lista) <~~> x" in subst, assumption);
apply(drule_tac z1=aa in cons_perm_eq[THEN iffD1]);
apply(assumption);
apply(subgoal_tac "aa mem (aa#lista)");
apply(drule sublistD, assumption);
apply(simp add: set_mem_eq);
apply(simp (no_asm));
apply(clarify);
apply(simp (no_asm) add: f_rem_def);
apply(fold f_rem_def);
apply(frule_tac x=a and xs=list and ys="aa#lista" in ldiff21a_cps1, assumption);
apply(drule f_rem_red);
back;
apply(erule ssubst);
apply(simp (no_asm));
apply(drule_tac x="aa#lista" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(rule sublist_subset[THEN iffD2]);
apply(drule sublist_subset[THEN iffD1]);
apply(rule subsetI);
apply(simp);
apply(erule disjE);
apply(elim conjE);
apply(simp add: set_mem_eq);
apply(elim conjE);
apply(drule subsetD, assumption);
apply(simp add: set_mem_eq);
apply(erule disjE);
apply(simp add: set_mem_eq);
apply(assumption);
apply(assumption);
done



theorem dist_sublist_is_rem_sublist : " \<lbrakk> distinct l; distinct r; sublist l r \<rbrakk>
                                        \<Longrightarrow> rem_sublist l r "
apply(simp add: rem_sublist_def);
apply(rule exI);
apply(rule conjI);
prefer 2;
apply(erule f_rem_perm_dist[rule_format, OF conjI]);
apply(simp);
apply(erule ldiff2_sublist3);
done


theorem rem_map_eq : " inj f \<Longrightarrow> map f (rem x xs) = rem (f x) (map f xs) "
apply(induct_tac xs);
apply(simp);
apply(simp only: map.simps rem.simps);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(rule refl);
apply(rule impI);
apply(erule notE);
apply(erule injD, assumption);
apply(rule impI);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(rule impI);
by simp;

theorem rem_map_perm[rule_format] : " x mem xs \<longrightarrow> map f (rem x xs) <~~> rem (f x) (map f xs) "
apply(induct_tac xs);
apply(simp);
apply(rule impI);
apply(simp only: map.simps rem.simps);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(rule impI);
apply(drule mp, simp add: set_mem_eq);
apply(simp);
apply(split split_if_asm);
apply(simp);
apply(drule rem_perm);
apply(drule_tac f=f in map_perm[rule_format]);
apply(rule Permutation.trans);
prefer 2;
apply(assumption);
apply(simp);
apply(rule impI);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(rule impI);
apply(drule mp, simp add: set_mem_eq);
apply(simp);
done


theorem f_rem_map_perm[rule_format, OF conjI] : 
                    " \<forall> xs l. sublist xs l \<and> distinct xs 
                      \<longrightarrow> map f (f_rem xs l) <~~> f_rem (map f xs) (map f l) "
apply(intro allI);
apply(induct_tac xs);
apply(simp add: f_rem_def);
apply(rule impI);
apply(simp add: f_rem_def);
apply(fold f_rem_def);
apply(elim conjE);
apply(subgoal_tac "a mem l");
apply(subgoal_tac "\<not> a mem list");
apply(drule f_rem_red2);
apply(assumption);
apply(drule_tac f=f in rem_map_perm);
back;
apply(rule Permutation.trans);
apply(assumption);
apply(drule mp);
apply(rule sublist_subset[THEN iffD2]);
apply(drule sublist_subset[THEN iffD1]);
apply(rule subsetI);
apply(drule_tac c=x in subsetD, simp);
apply(assumption);
apply(rule rem_perm2);
apply(assumption);
apply(simp add: set_mem_eq);
apply(subgoal_tac "a mem (a#list)");
apply(drule sublistD, assumption);
apply(assumption);
by (simp (no_asm));



theorem rem_sublist_map : "\<lbrakk> distinct l; distinct r; sublist l r \<rbrakk> 
                           \<Longrightarrow> rem_sublist (map f l) (map f r) "
apply(frule f_rem_perm_dist[rule_format,OF conjI]);
apply(rule conjI);
apply(assumption);
back;
apply(assumption);
apply(frule ldiff2_dist);
back;
apply(simp add: rem_sublist_def);
apply(rule_tac x="map f (ldiff2 r l)" in exI);
apply(rule conjI);
apply(rule sublist_map);
apply(erule ldiff2_sublist3);
apply(rule Permutation.trans);
prefer 2;
apply(rule f_rem_map_perm);
apply(erule ldiff2_sublist3);
apply(assumption);
apply(erule map_perm[rule_format]);
done


theorem rem_sublist_sorted : " \<lbrakk> sorted le l; sorted le r; rem_sublist l r;
                                         (\<forall> x y. le x y \<and> le y x \<longrightarrow> x = y) \<rbrakk>
                               \<Longrightarrow> \<exists> xs. l = f_rem xs r "
apply(simp add: rem_sublist_def);
apply(clarify);
apply(drule_tac l=r and xs=xs in f_rem_sorted);
apply(rule_tac x=xs in exI);
apply(rule sorted_eq[rule_format]);
apply(drule spec);
apply(drule spec);
apply(drule mp, assumption);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
by assumption;






end