
theory Jobs = Greedy_Aux3:


types Job = " nat \<times> nat "

consts
Input :: "Job list"


constdefs
date :: "Job \<Rightarrow> nat"
"date \<equiv> fst"

val :: "Job \<Rightarrow> nat"
"val \<equiv> snd"

val_op :: "[nat, nat] \<Rightarrow> nat"
"val_op a b \<equiv> a + b"

val_ord :: "[nat, nat] \<Rightarrow> bool"
"val_ord a b \<equiv> b \<le> a"

val_unit :: nat
"val_unit \<equiv> 0"

Job_val_ord :: "[Job, Job] \<Rightarrow> bool"
"Job_val_ord \<equiv> \<lambda> j j'. val_ord (val j) (val j')"

Job_date_ord :: " Job \<Rightarrow> Job \<Rightarrow> bool "
"Job_date_ord x y \<equiv> date x \<le> date y"

axioms
 In1 : "distinct Input"
 In2 : "sorted Job_val_ord Input"






lemma val_op1 : " \<forall> x. val_op val_unit x = x "
apply(simp add: val_op_def val_unit_def);
done

lemma val_op2 : " \<forall> x y. val_op x y = val_op y x "
apply(simp add: val_op_def);
done

lemma val_op3 : " \<forall> x y z. val_op x (val_op y z) = val_op (val_op x y) z"
apply(simp add: val_op_def);
done


lemma val_ord_refl : "\<forall> x. val_ord x x"
by (simp add: val_ord_def);

lemma val_ord_trans : "\<forall> x y z. val_ord x y \<and> val_ord y z \<longrightarrow> val_ord x z"
apply (simp add: val_ord_def val_def);
apply(clarify);
by simp;

lemma val_ord_mono  : " \<forall> a b d. val_ord a b  \<longrightarrow> val_ord (val_op a d) (val_op b d) " 
apply (simp add: val_ord_def val_def val_op_def);
done

lemma val_ord_total : " \<forall> x y. val_ord x y \<or> x = y \<or> val_ord y x "
apply(clarify);
by (simp add: val_ord_def);




lemma date_ord_refl : " \<forall> x. Job_date_ord x x "
by(simp add: Job_date_ord_def);

lemma date_ord_trans : " \<forall> x y z. Job_date_ord x y \<longrightarrow> Job_date_ord y z \<longrightarrow> 
                                          Job_date_ord x z"
by(simp add: Job_date_ord_def);

lemma date_ord_total : " \<forall> x y. Job_date_ord x y \<or> x = y \<or> Job_date_ord y x "
apply(clarify);
by(simp add: Job_date_ord_def);




consts
sortf :: "Job list \<Rightarrow> Job list"
axioms
A_sort_date : " \<forall> j. (sorted Job_date_ord (sortf j)) \<and> (j <~~> (sortf j)) "
A_sort_val : " \<forall> j. (sorted Job_val_ord (sortf j)) \<and> (j <~~> (sortf j)) "


section{* sortf properties *}

lemma sorted_corr : " (sorted Job_date_ord l) = (sorted op \<le> (map date l)) "
apply(induct_tac l);
apply(simp);
apply(simp);
apply(rule iffI);
apply(rule conjI);
apply(clarify);
apply(drule_tac x="(aa, ba)" in bspec, assumption);
apply(simp add: Job_date_ord_def date_def);
apply(erule conjE);
apply(assumption);
apply(rule conjI);
apply(clarify);
apply(drule_tac x="(aa, ba)" in bspec, assumption);
apply(simp add: Job_date_ord_def date_def);
apply(erule conjE);
by(assumption);

lemma sort_date_eq : " \<forall> l r. l <~~> r \<longrightarrow> map date (sortf l) = map date (sortf r) "
apply(clarify);
apply(insert A_sort_date);
apply(frule_tac x="l" in spec);
apply(drule_tac x="r" in spec);
apply(elim conjE);
apply(drule Permutation.trans);
apply(assumption);
apply(drule perm_sym);
apply(drule Permutation.trans);
back;
back;
apply(assumption);
apply(rule_tac le="op \<le>" in sorted_eq[rule_format]);
apply(rule le_anti_sym);
apply(simp);
apply(simp);
apply(rule conjI);
apply(rule_tac f="date" in map_perm[rule_format]);
apply(assumption);
apply(rule conjI);
apply(rule sorted_corr[THEN iffD1]);
apply(assumption);
apply(rule sorted_corr[THEN iffD1]);
by assumption;


subsection{* Specifications of feasible solutions *}

consts
F_aux :: " nat \<Rightarrow> nat list \<Rightarrow> bool"
primrec 
"F_aux n [] = True"
"F_aux n (j#js) = (if (n \<le> j) then (F_aux (Suc n) js) else False) " 

constdefs
Feas :: "Job list \<Rightarrow> bool"
"Feas j \<equiv> distinct j \<and> sublist j Input \<and> F_aux 0 (map date (sortf j))"


subsubsection{* Properties of F_aux *}

theorem F_aux1a[rule_format] : " \<forall> n. (\<forall> j p. j mem l \<and> p mem pos n j l \<longrightarrow> p \<le> j) 
                                \<longrightarrow> F_aux n l"
apply(induct_tac l);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(rule conjI);
prefer 2;
apply(drule_tac x=a in spec);
back;
apply(drule_tac x=n in spec);
back;
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(simp);
apply(assumption);
apply(rule impI);
apply(drule_tac x="Suc n" in spec);
apply(drule mp);
apply(clarify);
apply(drule_tac x="j" in spec);
apply(drule_tac x=p in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(simp);
apply(clarify);
by assumption;

theorem F_aux1 : " \<forall> j. j mem l \<longrightarrow> last_ind j l \<le> j 
                   \<Longrightarrow> F_aux 0 l"
apply(rule F_aux1a);
apply(erule conjE);
apply(drule_tac x=j in spec);
apply(drule mp, assumption);
apply(drule last_ubnd);
apply(rule le_trans);
by assumption;


theorem F_aux2a[rule_format] : " \<forall> n. F_aux n l \<longrightarrow> 
                                  (\<forall> j p. j mem l \<and> p mem pos n j l \<longrightarrow> p \<le> j)"
apply(induct_tac l);
apply(clarify);
apply(simp);
apply(clarify);
apply(case_tac "a = j");                        
apply(clarify);
apply(simp only: pos.simps);
apply(split split_if_asm);
apply(simp only: F_aux.simps);
apply(split split_if_asm);
apply(case_tac "n = p");
apply(simp);
apply(simp);
apply(case_tac "j mem list");
apply(drule_tac x="Suc n" in spec);
apply(drule mp, assumption);
apply(drule_tac x="j" in spec);
apply(drule_tac x="p" in spec);
apply(drule mp, simp);
apply(assumption);
apply(drule_tac n1="Suc n" in pos2[rule_format, THEN[2] contrapos_nn], assumption);
apply(simp);
apply(simp);
apply(simp);
apply(simp only: pos.simps);
apply(split split_if_asm);
apply(simp);
apply(simp only: F_aux.simps);
apply(split split_if_asm);
apply(drule_tac x="Suc n" in spec);
apply(drule mp, assumption);
apply(drule_tac x="j" in spec);
apply(drule_tac x="p" in spec);
apply(drule mp, simp);
apply(assumption);
by simp;

theorem F_aux2 : " \<lbrakk> F_aux 0 l; j mem l \<rbrakk>
                   \<Longrightarrow> last_ind j l \<le> j "
apply(erule F_aux2a);
apply(rule conjI);
apply(assumption);
apply(erule last_mem);
done


subsubsection {* Properties of F_aux2 *} 

consts
F_aux2 :: " nat \<Rightarrow> Job list \<Rightarrow> bool"
primrec 
"F_aux2 n [] = True"
"F_aux2 n (j#js) = (if (n \<le> (date j)) then (F_aux2 (Suc n) js) else False) " 


theorem F_aux21a[rule_format] : " \<forall> n. 
                                 (\<forall> j p. j mem js \<and> p mem pos n j js \<longrightarrow> p \<le> date j) 
                                \<longrightarrow> F_aux2 n js"
apply(induct_tac js);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(rule conjI);
prefer 2;
apply(drule_tac x="(a, b)" in spec);
apply(drule_tac x=n in spec);
back;
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(simp);
apply(assumption);
apply(rule impI);
apply(drule_tac x="Suc n" in spec);
apply(drule mp);
apply(clarify);
apply(drule_tac x="(aa, ba)" in spec);
apply(drule_tac x=p in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(simp);
apply(clarify);
by assumption;

theorem F_aux21 : " \<forall> j. j mem js \<longrightarrow> last_ind j js \<le> date j 
                   \<Longrightarrow> F_aux2 0 js"
apply(rule F_aux21a);
apply(erule conjE);
apply(drule_tac x=j in spec);
apply(drule mp, assumption);
apply(drule last_ubnd);
apply(rule le_trans);
by assumption;


theorem F_aux22a[rule_format] : " \<forall> n. F_aux2 n js \<longrightarrow> 
                                  (\<forall> j p. j mem js \<and> p mem pos n j js \<longrightarrow> p \<le> date j)"
apply(induct_tac js);
apply(clarify);
apply(simp);
apply(clarify);
apply(case_tac "(a, b) = (aa, ba)");                        
apply(clarify);
apply(simp only: pos.simps);
apply(split split_if_asm);
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(case_tac "n = p");
apply(simp);
apply(simp);
apply(case_tac "(aa, ba) mem list");
apply(drule_tac x="Suc n" in spec);
apply(drule mp, assumption);
apply(drule_tac x="aa" in spec);
apply(drule_tac x="ba" in spec);
apply(drule_tac x="p" in spec);
apply(drule mp, simp);
apply(assumption);
apply(drule_tac n1="Suc n" in pos2[rule_format, THEN[2] contrapos_nn], assumption);
apply(simp);
apply(simp);
apply(simp);
apply(simp only: pos.simps);
apply(split split_if_asm);
apply(simp);
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(drule_tac x="Suc n" in spec);
apply(drule mp, assumption);
apply(drule_tac x="(aa, ba)" in spec);
apply(drule_tac x="p" in spec);
apply(drule mp, simp);
apply(assumption);
by simp;

theorem F_aux22 : " \<lbrakk> F_aux2 0 js; j mem js \<rbrakk>
                   \<Longrightarrow> last_ind j js \<le> date j "
apply(erule F_aux22a);
apply(rule conjI);
apply(assumption);
apply(erule last_mem);
done



lemma F_aux23[rule_format] : " \<forall> n. F_aux2 (Suc n) j \<longrightarrow> F_aux2 n j "
apply(induct_tac j);
apply(simp);
apply(rule allI);
apply(rule impI);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(drule_tac x="Suc n" in spec);
apply(simp);
apply(split split_if_asm);
apply(simp);
apply(simp);
apply(simp);
apply(split split_if_asm);
apply(drule spec);
apply(simp);
apply(simp);
done


lemma F_aux24 : " \<forall> m n. F_aux2 (n + m) j \<longrightarrow> 
                  (\<forall> d. d \<le> m \<longrightarrow> F_aux2 (n + d) j) "
apply(rule allI);
apply(induct_tac m);
apply(simp);
apply(rule allI);
apply(rule impI);
apply(rule allI);
apply(rule impI);
apply(case_tac "d = Suc n");
apply(simp);
apply(subgoal_tac "na + Suc n = Suc (na + n)");
apply(drule subst);
apply(assumption);
back;
apply(drule_tac n="na + n" in F_aux23);
apply(drule_tac x="na" in spec);
apply(drule mp, assumption);
apply(drule_tac x="d" in spec);
apply(drule mp, simp);
apply(assumption);
by simp;

lemma F_aux25_a[rule_format] : "\<forall> n.  F_aux2 n (js @ js') \<longrightarrow> F_aux2 n js "
apply(induct_tac js);
apply(simp);
apply(clarify);
apply(simp);
apply(split split_if_asm);
apply(assumption);
by simp;

lemma F_aux25_b[rule_format] : "\<forall> n.  F_aux2 n (js @ js') \<longrightarrow> F_aux2 (n + (length js)) js' "
apply(induct_tac js);
apply(simp);
apply(rule allI);
apply(rule impI);
apply(simp);
apply(split split_if_asm);
apply(drule_tac x="Suc n" in spec);
apply(drule mp, assumption);
by simp_all;


lemma F_aux25_c[rule_format, OF conjI] : 
             "\<forall> n js'. F_aux2 n js \<and> F_aux2 (n + length js) js' \<longrightarrow> F_aux2 n (js @ js') "
apply(induct_tac js);
apply(simp);
apply(intro allI);
apply(rule impI);
apply(erule conjE);
apply(simp only: F_aux2.simps);
apply(simp (no_asm));
apply(split split_if_asm);
apply(rule conjI);
apply(rule impI);
apply(drule_tac x="Suc n" in spec);
apply(drule_tac x="js'" in spec);
apply(drule mp, simp);
apply(assumption);
apply(assumption);
by simp;




lemma F_aux2_eqa : " \<forall> n. F_aux n (map date js) \<longrightarrow> F_aux2 n js " 
apply(induct_tac js);
apply(simp);
apply(clarify);
apply(simp);
apply(split split_if_asm);
by simp_all;

lemma F_aux2_eqb : " \<forall> n. F_aux2 n js \<longrightarrow> F_aux n (map date js) " 
apply(induct_tac js);
apply(simp);
apply(clarify);
apply(simp);
apply(split split_if_asm);
by simp_all;

lemma F_aux2_eq : " F_aux n (map date js) = F_aux2 n js " 
apply(rule iffI);
apply(erule F_aux2_eqa[rule_format]);
apply(erule F_aux2_eqb[rule_format]);
done

constdefs
Feas' :: "Job list \<Rightarrow> bool"
"Feas' j \<equiv> distinct j \<and> sublist j Input \<and> F_aux2 0 (sortf j)"

lemma Feas_eq : " Feas j = Feas' j "
apply(simp add: Feas_def Feas'_def);
apply(rule iffI);
apply(simp);
apply(elim conjE);
apply(erule F_aux2_eq[THEN iffD1]);
apply(simp);
apply(elim conjE);
apply(erule F_aux2_eq[THEN iffD2]);
done



lemma F_aux2_subst : "\<forall> n.  F_aux2 n js \<and> date a = date b \<longrightarrow> F_aux2 n (subst b a js) "
apply(induct_tac js);
apply(simp);
apply(rule allI);
apply(rule impI);
apply(erule conjE);
apply(simp only: subst.simps);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(rule F_aux2_eq[THEN iffD1]);
apply(drule F_aux2_eq[THEN iffD2]);
apply(simp);
apply(rule impI);
apply(simp (no_asm));
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(rule conjI);
apply(rule impI);
apply(drule_tac x="Suc n" in spec);
apply(drule mp, simp);
by simp_all;

lemma F_aux2_hd_swap[rule_format, OF conjI] : 
                       "\<forall> n. F_aux2 n (j#js) \<and> j' mem js \<and> date j' < date j
                        \<longrightarrow> F_aux2 n (j'#(subst j j' js)) "
apply(induct_tac js);
apply(simp);
apply(rule allI);
apply(rule impI);
apply(elim conjE);
apply(simp only: subst.simps);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(split split_if_asm);
apply(split split_if_asm);
apply(split split_if_asm);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(assumption);
apply(rule impI);
apply(drule not_leE);
apply(drule_tac j="date j" in less_trans, assumption);
apply(drule_tac j="date j'" in le_less_trans, assumption);
apply(simp);
apply(rule impI);
apply(simp);
apply(simp);
apply(simp);
apply(rule impI);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(rule conjI);
apply(rule impI);
apply(drule_tac x="Suc n" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(simp);
apply(simp);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(drule_tac x=n in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(split split_if_asm);
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(drule F_aux23);
apply(assumption);
apply(assumption);
apply(simp);
apply(simp);
apply(simp);
apply(split split_if_asm);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(simp);
apply(simp);
apply(split split_if_asm);
apply(split split_if_asm);
apply(split split_if_asm);
apply(simp);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(simp);
done


lemma F_aux2_sort[rule_format, OF conjI] : " \<forall> js n. F_aux2 n js \<and> js <~~> js' \<and> sorted Job_date_ord js'
                                             \<longrightarrow> F_aux2 n js' "
apply(induct_tac js');
apply(simp);
apply(intro allI);
apply(rule impI);
apply(elim conjE);
apply(case_tac js);
apply(simp);
apply(case_tac "aa = a");
apply(drule_tac x=lista in spec);
apply(drule_tac x="Suc n" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(simp);
apply(simp);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(subgoal_tac "aa \<in> set list");
prefer 2;
apply(subgoal_tac "aa mem js");
apply(drule perm_mem, assumption);
apply(simp add: set_mem_eq);
apply(erule ssubst);
apply(simp (no_asm));
apply(simp only: sorted.simps);
apply(erule conjE);
apply(drule_tac x=aa in bspec, assumption);
apply(simp only: Job_date_ord_def);
apply(subgoal_tac "a mem lista");
prefer 2;
apply(subgoal_tac "a mem (a#list)");
apply(drule perm_sym);
apply(drule perm_mem, assumption);
apply(simp add: set_mem_eq);
apply(erule disjE);
apply(simp);
apply(assumption);
apply(simp (no_asm));
apply(case_tac "date a < date aa");
apply(frule_tac y=a and x=aa in hd_swap[rule_format, OF conjI]);
apply(clarify);
apply(drule_tac j'=a in F_aux2_hd_swap);
apply(simp);
apply(drule_tac x="subst aa a lista" in spec);
apply(drule_tac x="Suc n" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(rule conjI);
apply(drule_tac ys="aa # lista" in Permutation.trans, assumption);
apply(simp);
apply(simp);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(assumption);
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(subgoal_tac "date a = date aa");
apply(frule_tac y=a and x=aa in hd_swap[rule_format, OF conjI]);
apply(rule notI);
apply(simp);
apply(drule_tac x="subst aa a lista" in spec);
apply(drule_tac x="Suc n" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(erule F_aux2_subst[rule_format, OF conjI]);
apply(assumption);
apply(simp);
apply(rule conjI);
apply(drule_tac ys="aa # lista" in Permutation.trans, assumption);
apply(simp);
apply(simp);
apply(simp (no_asm));
apply(rule conjI);
apply(rule impI);
apply(assumption);
apply(erule ssubst);
back;
apply(simp only: F_aux2.simps);
apply(split split_if_asm);
apply(assumption);
apply(simp);
apply(simp);
done


lemma F_aux2_f_rem : " F_aux2 0 js \<Longrightarrow> F_aux2 0 (f_rem js' js) "
apply(rule F_aux21);
apply(rule allI);
apply(rule impI);
apply(frule last_f_rem_pos);
apply(frule f_rem_mem);
apply(drule F_aux22, assumption);
by simp;





subsection {* Properties of Feas *}


theorem FA1 : "\<forall> A. Feas A \<longrightarrow> (sublist A Input) \<and> (distinct A) "
apply(clarify);
apply(simp add: Feas_def);
done

theorem FA2 : "\<forall> A B. Feas A \<and> A <~~> B \<longrightarrow> Feas B "
apply(clarify);
apply(simp add: Feas_def);
apply(rule conjI);
apply(clarify);
apply(erule perm_dist);
apply(assumption);
apply(elim conjE);
apply(frule sort_date_eq[rule_format]);
apply(rule conjI);
apply(erule sublist_perm, assumption);
by simp;


theorem M1 : " Feas [] "
apply(simp add: Feas_def);
apply(insert A_sort_date);
apply(drule_tac x="[]" in spec);
apply(erule conjE);
apply(rule conjI);
apply(simp add: sublist_def);
by simp;

theorem M2 : " \<forall> A B. distinct A \<and> Feas B \<and> sublist A B \<longrightarrow> Feas A "
apply(clarify);
apply(simp add: Feas_def);
apply(elim conjE);
apply(rule conjI);
apply(rule sublist_trans);
apply(assumption);
apply(assumption);
apply(insert A_sort_date);
apply(frule_tac x="A" in spec);
apply(drule_tac x="B" in spec);
apply(clarify);
apply(drule_tac l=A in perm_dist, assumption);
apply(drule_tac l=B in perm_dist, assumption);
apply(subgoal_tac "sublist (sortf A) (sortf B)"); 
prefer 2;
apply(drule sublist_perm3);
apply(drule sublist_perm3);
apply(elim conjE);
apply(simp add: sublist_subset);
apply(drule_tac l="sortf A" and f="date" in rem_sublist_map);
apply(assumption);
apply(assumption);
apply(drule sorted_corr[THEN iffD1]);
apply(drule sorted_corr[THEN iffD1]);
apply(drule_tac xs="sortf A" and f="date" in sublist_map);
apply(drule_tac l="map date (sortf A)" in rem_sublist_sorted);
apply(assumption);
apply(assumption);
apply(clarify);
apply(rule le_anti_sym);
apply(assumption);
apply(assumption);
apply(erule exE);
apply(rule F_aux1);
apply(clarify);
apply(drule rem_sublist_is_sublist);
apply(drule_tac x=j in sublistD, assumption);
apply(drule F_aux2);
apply(assumption);
apply(frule_tac P="\<lambda> x. j mem x" in subst);
apply(assumption);
apply(drule last_f_rem_pos);
apply(erule ssubst);
apply(rule le_trans);
apply(assumption);
apply(assumption);
done







consts
sel_max :: "Job list \<Rightarrow> Job list \<Rightarrow> (Job option \<times> Job list)"
primrec
"sel_max l [] = (None, [])"
"sel_max l (x#xs) = (if (x mem l) then (let (opt, jlst) = sel_max (rem x l) xs 
                                        in
                                         (opt, jlst @ [x]))
                     else (Some x, []))"


lemma sel_max0 : " \<forall> l. (\<exists> x. x mem r \<and> (\<not> x mem l)) 
                   \<longrightarrow> (\<exists> x js. sel_max l r = (Some x, js)) "
apply(induct_tac r);
apply(simp);
apply(clarify);
apply(simp (no_asm));
apply(rule impI);
apply(case_tac "aa = a");
apply(case_tac "ba = b");
apply(simp);
apply(drule_tac x="rem (a, b) l" in spec);
apply(drule mp);
apply(rule_tac x="(aa, ba)" in exI);
apply(rule conjI);
apply(simp);
apply(rule notI);
apply(erule notE);
apply(drule rem_aux);
apply(assumption);
apply(elim exE);
apply(simp add: Let_def);
apply(drule_tac x="rem (a, b) l" in spec);
apply(drule mp);
apply(rule_tac x="(aa, ba)" in exI);
apply(rule conjI);
apply(simp);
apply(rule notI);
apply(erule notE);
apply(drule rem_aux);
apply(assumption);
apply(elim exE);
apply(simp add: Let_def);
done

lemma sel_max1 : " \<lbrakk> distinct r; length l < length r \<rbrakk>
                   \<Longrightarrow> (\<exists> x js. sel_max l r = (Some x, js)) "
apply(drule sublist_length1);
apply(assumption);
apply(simp add: sublist_def);
apply(drule sel_max0[rule_format, simplified]);
by assumption;

lemma sel_max2a[rule_format] : "\<forall> l. x \<in> set (snd (sel_max l r)) \<longrightarrow> (x mem l) \<and> (x mem r) "
apply(induct_tac r);
apply(simp);
apply(rule allI);
apply(rule impI);
apply(simp);
apply(split split_if_asm);
apply(rule conjI);
apply(clarify);
apply(clarify);
apply(simp add: Let_def);
apply(simp add: split_def);
apply(erule disjE);
apply(simp);
apply(drule_tac x="rem (a, b) l" in spec);
apply(drule mp, assumption);
apply(erule conjE);
apply(drule rem_aux);
apply(simp);
by simp;

lemma sel_max2b : " sel_max l r = (Some x, js) \<Longrightarrow> sublist js r \<and> sublist js l"
apply(rule conjI);
apply(simp add: sublist_def);
apply(clarify);
apply(rule_tac l1=l in sel_max2a[THEN conjunct2]);
apply(simp add: set_mem_eq);
apply(simp add: sublist_def);
apply(clarify);
apply(rule_tac r1=r in sel_max2a[THEN conjunct1]);
apply(simp add: set_mem_eq);
done


lemma sel_max2c[rule_format, OF conjI] : 
        " \<forall> l x js. distinct r \<and> sel_max l r = (Some x, js) \<longrightarrow> distinct js "
apply(induct_tac r);
apply(simp);
apply(intro allI);
apply(rule impI);
apply(elim conjE);
apply(subgoal_tac "sel_max l (a # list) = ?y");
prefer 2;
apply(rule sel_max.simps);
apply(split split_if_asm);
apply(drule_tac x="rem a l" in spec);
apply(drule_tac x=x in spec);
apply(drule_tac x="rev (tl (rev js))" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp);
apply(simp add: Let_def split_def);
apply(elim conjE);
apply(erule subst);
apply(erule subst);
apply(simp);
apply(simp add: Let_def split_def);
apply(elim conjE);
apply(frule_tac s=js and P="\<lambda> x. distinct (?f x)" in ssubst); 
apply(assumption);
apply(erule_tac t=js in subst);
apply(simp);
apply(rule notI);
apply(drule sel_max2a);
apply(simp add: set_mem_eq);
apply(drule spec);
apply(drule spec);
apply(drule spec);
apply(simp);
done




lemma sel_max3[rule_format] : "\<forall> l. sorted le r \<longrightarrow> sorted le (snd (sel_max l (rev r))) "
apply(induct_tac r rule: rev_induct);
apply(simp);
apply(rule allI);
apply(rule impI);
apply(simp);
apply(clarify);
apply(simp (no_asm) add: Let_def);
apply(simp (no_asm) add: split_def);
apply(rule conjI);
apply(drule_tac x="rem (a, b) l" in spec);
apply(assumption);
apply(drule_tac x="rem (a, b) l" in spec);
apply(clarify);
apply(drule sel_max2a);
apply(erule conjE);
apply(drule_tac x="(aa, ba)" in bspec);
apply(drule rev_mem);
apply(simp add: set_mem_eq);
by assumption;


lemma sel_max4[rule_format, OF conjI] : " \<forall> l x js. distinct r \<and> sel_max l r = (Some x, js) 
                                          \<longrightarrow> (x mem r \<and> (\<not> x mem l)) "
apply(induct_tac r);
apply(simp);
apply(clarify);
apply(simp);
apply(split split_if_asm);
apply(rule conjI);
apply(clarify);
apply(simp add: Let_def split_def);
apply(clarify);
apply(drule_tac x="rem (aa, ba) l" in spec);
apply(drule_tac x=aa in spec);
apply(drule_tac x=ba in spec);
apply(drule mp);
apply(rule_tac x="snd (sel_max (rem (aa, ba) l) list)" in exI);
apply(erule subst);
apply(simp);
apply(simp add: set_mem_eq);
apply(clarify);
apply(simp add: Let_def split_def);
apply(clarify);
apply(drule_tac x="rem (a, b) l" in spec);
apply(drule_tac x=aa in spec);
apply(drule_tac x=ba in spec);
apply(drule mp);
back;
apply(rule_tac x="snd (sel_max (rem (a, b) l) list)" in exI);
apply(erule subst);
apply(simp);
apply(clarify);
apply(subgoal_tac "(a, b) \<noteq> (aa, ba)");
apply(drule rem_aux2);
apply(assumption);
back;
apply(simp);
apply(clarify);
apply(simp);
by simp;

lemma dist_rev[rule_format] : " distinct l \<longrightarrow> distinct (rev l) "
apply(induct_tac l rule: rev_induct);
by simp_all;



lemma sel_max5a[rule_format, OF conjI] : "\<forall> l x js. distinct r \<and>
                                                          sorted le r \<and> sel_max l (rev r) = (Some x, js) 
                                         \<longrightarrow> (\<forall> y. y mem js \<longrightarrow> le x y) "
apply(induct_tac r rule: rev_induct);
apply(simp);
apply(clarify);
apply(simp);
apply(clarify);
apply(split split_if_asm);
apply(simp add: Let_def split_def);
apply(erule conjE);
apply(case_tac "(ab, bb) = (a, b)");
apply(clarify);
apply(drule_tac x="(aa, ba)" in bspec);
apply(rule set_mem_eq[THEN iffD1]);
apply(rule rev_mem);
apply(rule_tac l1="rem (a, b) l" and js1="snd (sel_max (rem (a, b) l) (rev xs))" in sel_max4[THEN conjunct1]);
apply(erule dist_rev);
apply(erule subst);
apply(simp);
apply(assumption);
apply(subgoal_tac "(ab, bb) mem (snd (sel_max (rem (a, b) l) (rev xs)))");
apply(drule_tac x="rem (a, b) l" in spec);
apply(drule_tac x=aa in spec);
apply(drule_tac x=ba in spec);
apply(drule_tac x="rev (tl (rev js))" in spec);
apply(drule mp);
apply(erule subst);
apply(erule subst);
apply(simp);
apply(drule_tac x=ab in spec);
apply(drule_tac x=bb in spec);
apply(drule mp);
apply(erule subst);
back;
apply(simp);
apply(assumption);
apply(drule_tac s=js and P="\<lambda> x. ?y mem x" in ssubst, assumption);
apply(simp add: set_mem_eq);
apply(erule disjE);
apply(clarify);
apply(simp);
apply(assumption);
by simp;

lemma sel_max5b[rule_format, OF conjI] : " \<forall> l x js y. sorted le r \<and> sel_max l (rev r) = (Some x, js) 
                                            \<longrightarrow> (y mem r \<and> (\<not> y mem l))
                                             \<longrightarrow> (le y x \<or> x = y) "
apply(induct_tac r rule: rev_induct);
apply(simp);
apply(intro allI);
apply(intro impI);
apply(elim conjE);
apply(rule disjCI);
apply(simp);
apply(split split_if_asm);
apply(simp add: Let_def split_def);
apply(erule conjE);
apply(case_tac "y = x");
apply(clarify);
apply(drule_tac x="rem x l" in spec);
apply(drule_tac x="fst xa" in spec);
apply(drule_tac x="snd xa" in spec);
apply(drule mp);
apply(rule_tac x="snd (sel_max (rem x l) (rev xs))" in exI);
apply(simp);
apply(erule subst);
apply(simp);
apply(drule_tac x="fst y" in spec);
apply(drule_tac x="snd y" in spec);
apply(drule mp);
apply(rule conjI);
apply(simp add: set_mem_eq);
apply(simp);
apply(rule notI);
apply(erule notE);
apply(erule rem_aux);
apply(erule disjE);
apply(simp);
apply(clarify);
apply(simp);
apply(simp);
apply(erule conjE);
apply(drule_tac x=y in bspec);
apply(simp add: set_mem_eq);
apply(erule disjE);
apply(simp);
apply(assumption);
apply(assumption);
done

lemma sel_max5c[rule_format, OF conjI] : "\<forall> l x js. distinct l \<and> distinct r \<and> sel_max l r = (Some x, js) 
                                         \<longrightarrow> (\<forall> y. y mem js \<longrightarrow> x \<noteq> y) "
apply(induct_tac r);
apply(simp);
apply(intro allI);
apply(rule impI);
apply(rule allI);
apply(rule impI);
apply(erule conjE);
apply(simp);
apply(erule conjE);
apply(split split_if_asm);
apply(simp add: Let_def split_def);
apply(elim conjE);
apply(case_tac "y = a");
apply(rule notI);
apply(drule_tac r=list and l="rem a l" and x=x and js="snd (sel_max (rem a l) list)" in sel_max4);
apply(erule_tac t="Some x" in subst);
apply(simp);
apply(erule conjE);
apply(simp add: set_mem_eq);
apply(drule_tac x="rem a l" in spec);
apply(drule_tac x="fst x" in spec);
apply(drule_tac x="snd x" in spec);
apply(drule_tac x="rev (tl (rev js))" in spec);
apply(drule mp);
apply(rule conjI);
apply(erule rem_dist3);
apply(assumption);
apply(simp);
apply(erule subst);
apply(erule subst);
apply(simp);
apply(drule_tac x="snd x" in spec);
apply(rule notI);
apply(drule mp);
apply(rule subst, assumption);
apply(simp (no_asm));
apply(simp add: set_mem_eq);
apply(drule_tac s=js in ssubst);
apply(assumption);
by simp_all;


lemma Suc_diff : " a \<le> b \<Longrightarrow> Suc a - Suc b = a - b "
by simp;

lemma Suc_diff2 : " Suc n' = n \<Longrightarrow> n + a - n' = Suc a "
apply(rule HOL.trans);
apply(rule diff_add_assoc2);
apply(simp);
apply(erule subst);
by simp;

lemma sel_max_F_aux[rule_format, OF conjI]
                  : " \<forall> l x js. distinct l \<and> distinct r  
                                         \<and> sel_max l (rev r) = (Some x, js) \<and> length l < length r
                                         \<and> F_aux2 0 r
                         \<longrightarrow> F_aux2 (Suc (length (f_rem js l))) js "
apply(induct_tac r rule: rev_induct);
apply(simp);
apply(intro allI);
apply(rule impI);
apply(elim conjE);
apply(simp);
apply(elim conjE);
apply(split split_if_asm);
apply(simp add: Let_def split_def);
apply(erule conjE);
apply(frule_tac j=x in F_aux22);
apply(simp (no_asm));
apply(simp only: last_of_last);
apply(subgoal_tac "x \<notin> set (snd (sel_max (rem x l) (rev xs)))");
prefer 2;
apply(rule notI);
apply(drule sel_max2a);
apply(simp add: set_mem_eq);
apply(drule_tac x="rem x l" in spec);
apply(drule_tac x="fst xa" in spec);
apply(drule_tac x="snd xa" in spec);
apply(drule_tac x="rev (tl (rev js))" in spec);
apply(drule mp);
apply(rule conjI);
apply(erule rem_dist3, assumption);
apply(rule conjI);
apply(erule_tac t=js in subst);
apply(simp);
apply(erule subst);
apply(simp);
apply(rule conjI);
apply(drule rem_length);
apply(simp);
apply(drule F_aux25_a);
apply(assumption);
apply(subgoal_tac "distinct js");
prefer 2;
apply(drule dist_rev);
apply(rule_tac t=js in subst, assumption);
apply(simp (no_asm));
apply(rule conjI);
apply(erule_tac r="rev xs" and l="rem x l" and x="xa" in sel_max2c);
apply(erule subst);
apply(simp);
apply(assumption);
apply(subgoal_tac "sublist js l");
prefer 2;
apply(simp only: sublist_def);
apply(intro allI);
apply(rule impI);
apply(case_tac "xb = x");
apply(simp);
apply(subgoal_tac "sel_max (rem x l) (rev xs) = (Some xa, snd (sel_max (rem x l) (rev xs))) ");
prefer 2;
apply(erule subst);
apply(simp);
apply(drule sel_max2b);
apply(erule conjE);
apply(simp only: sublist_def);
apply(drule_tac x=xb in spec);
back;
apply(drule mp);
apply(simp add: set_mem_eq);
apply(drule_tac s=js and P="\<lambda> x. xb \<in> set x" in ssubst, assumption);
apply(simp);
apply(erule rem_aux);
apply(frule_tac xs=js in f_rem_length2, assumption);
apply(frule rem_length);
apply(elim conjE);
apply(subgoal_tac "Suc (length (snd (sel_max (rem x l) (rev xs)))) = length js");
prefer 2;
apply(erule_tac t=js in subst);
apply(simp);
apply(subgoal_tac "length (snd (sel_max (rem x l) (rev xs))) \<le> (length (rem x l))");
prefer 2;
apply(simp);
apply(erule_tac t="length l" in subst);
apply(erule_tac t="length js" in subst);
apply(drule Suc_diff);
back;
back;
apply(simp);
apply(rule F_aux21a);
apply(erule conjE);
apply(subgoal_tac "F_aux2 (Suc (length (f_rem (snd (sel_max (rem x l) (rev xs))) (rem x l))))
(snd (sel_max (rem x l) (rev xs)))");
prefer 2;
apply(drule_tac s=js and P="\<lambda> x. F_aux2 (?f x) (?g x)" in ssubst);
apply(assumption);
back;
apply(simp);
apply(subgoal_tac "distinct (snd (sel_max (rem x l) (rev xs)))");
prefer 2;
apply(drule_tac s=js and P="\<lambda> x. distinct x" in ssubst, assumption);
apply(simp);
apply(subgoal_tac "sublist (snd (sel_max (rem x l) (rev xs))) (rem x l)");
prefer 2;
apply(simp only: sublist_def);
apply(rule allI);
apply(rule impI);
apply(case_tac "xb = x");
apply(simp add: set_mem_eq);
apply(drule_tac x=xb in spec);
apply(drule mp);
apply(erule_tac t=js in subst);
apply(simp add: set_mem_eq);
apply(rule_tac x=x in rem_aux2);
apply(rule notI);
apply(simp);
apply(assumption);
apply(drule_tac xs="snd (sel_max (rem x l) (rev xs))" in f_rem_length2);
apply(assumption);
apply(case_tac "j = x");
apply(frule_tac s=js and P="\<lambda> x. p mem ?f x" in ssubst);
apply(assumption);
apply(drule pos4[rule_format]);
apply(drule_tac y=js and f=length in arg_cong);
apply(simp);
apply(drule rem_length);
apply(simp);
apply(elim conjE);
apply(frule_tac s=js and P="\<lambda> x. p mem ?f x" in ssubst);
apply(assumption);
apply(drule_tac j=j and p=p in F_aux22a);
back;
apply(rule conjI);
apply(rule_tac t=js in subst, assumption);
apply(simp (no_asm));
apply(drule_tac s=js and P="\<lambda> x. j mem x" in ssubst, assumption);
apply(simp add: set_mem_eq);
apply(rule_tac t=js in subst, assumption);
apply(simp (no_asm));
apply(erule_tac t="length (f_rem (snd (sel_max (rem x l) (rev xs))) (rem x l))" in ssubst);
apply(rule_tac P="p mem ?F" in mp);
prefer 2;
apply(assumption);
back;
apply(simp (no_asm) add: pos_append);
apply(rule conjI);
apply(rule impI);
apply(simp);
apply(rule impI);
apply(simp);
apply(erule_tac t="length l" in subst);
apply(simp add: Suc_diff);
apply(assumption);
apply(simp);
done


lemma sel_max_F_aux2[rule_format, OF conjI] : 
                     " \<forall> l x js. distinct l \<and> distinct r 
                                         \<and> sel_max l (rev r) = (Some x, js) \<and> length l < length r
                                         \<and> F_aux2 0 l \<and> F_aux2 0 r
                         \<longrightarrow> F_aux2 0 ((f_rem js l)@[x]) "
apply(induct_tac r rule: rev_induct);
apply(simp);
apply(intro allI);
apply(rule impI);
apply(elim conjE);
apply(simp);
apply(elim conjE);
apply(split split_if_asm);
apply(simp add: Let_def split_def);
apply(elim conjE);
apply(rule_tac t=js in subst);
apply(assumption);
apply(simp (no_asm) add: f_rem_def);
apply(fold f_rem_def);
apply(drule_tac x="rem x l" in spec);
apply(drule_tac x="fst xa" in spec);
apply(drule_tac x="snd xa" in spec);
apply(drule_tac x="rev(tl (rev js))" in spec);
apply(drule mp);
apply(rule conjI);
apply(erule rem_dist3, assumption);
apply(rule conjI);
apply(erule_tac t=js in subst);
apply(simp (no_asm));
apply(erule subst);
apply(simp);
apply(rule conjI);
apply(drule rem_length);
apply(simp);
apply(rule conjI);
apply(drule_tac js'="[x]" in F_aux2_f_rem);
apply(simp add: f_rem_def);
apply(drule F_aux25_a);
apply(assumption);
apply(simp);
apply(drule_tac s=js and P="\<lambda> x. F_aux2 0 (?f x)" in ssubst);
apply(assumption);
back;
back;
apply(simp);
apply(simp add: f_rem_def);
apply(rule F_aux21);
apply(rule allI);
apply(rule impI);
apply(case_tac "j = xa");
apply(simp add: last_of_last);
apply(drule_tac j=xa in F_aux22);
back;
apply(simp (no_asm));
apply(simp add: last_of_last);
apply(subgoal_tac "j mem l");
apply(drule_tac j=j in F_aux22);
apply(assumption);
apply(drule_tac x=j in last_of_last_neq2);
back;
apply(assumption);
apply(simp);
apply(simp add: set_mem_eq);
done


lemma sel_max_perm : " \<lbrakk> distinct r; sel_max l r = (Some x, js) \<rbrakk> 
                         \<Longrightarrow> (l @ [x]) <~~> ((f_rem js l)@[x])@js "
apply(frule sel_max2b);
apply(erule conjE);
apply(frule sel_max2c);
apply(assumption);
apply(drule f_rem_perm3);
back;
apply(assumption);
back;
apply(simp);
apply(rule Permutation.trans);
apply(rule perm_append_swap);
apply(rule Permutation.trans);
prefer 2;
apply(rule perm_append_Cons);
apply(simp);
apply(rule Permutation.trans);
prefer 2;
apply(rule perm_append_swap);
by assumption;



theorem M3 : " \<forall> A B. Feas A \<and> Feas B \<and> (length A < length B) 
                       \<longrightarrow> (\<exists> e. (e mem B) \<and> (\<not> (e mem A)) 
                                          \<and> (Feas (A @ [e]))) "
apply(intro allI);
apply(rule impI);
apply(elim conjE);
apply(drule Feas_eq[THEN iffD1]);
apply(drule Feas_eq[THEN iffD1]);
apply(simp add: Feas'_def);
apply(elim conjE);
apply(insert A_sort_date);
apply(frule_tac x=A in spec);
apply(drule_tac x=B in spec);
apply(elim conjE);
apply(frule_tac l=A in perm_dist, assumption);
apply(frule_tac l=B in perm_dist, assumption);
apply(subgoal_tac "length (sortf A) < length (sortf B)");
prefer 2;
apply(drule perm_length);
apply(drule perm_length);
apply(simp);
apply(frule_tac l="sortf B" in dist_rev);
apply(subgoal_tac "length (sortf A) < length (rev(sortf B))");
prefer 2;
apply(simp);
apply(frule_tac r="rev(sortf B)" in sel_max1);
apply(assumption);
apply(elim exE);
apply(frule_tac r="rev(sortf B)" in sel_max4);
apply(assumption);
apply(elim conjE);
apply(subgoal_tac "x mem B");
prefer 2;
apply(rule perm_mem);
apply(rule perm_sym);
apply(assumption);
apply(rule perm_mem);
apply(rule perm_rev);
apply(simp);
apply(subgoal_tac "\<not> (x mem A)");
prefer 2;
apply(rule notI);
apply(erule notE);
apply(rule perm_mem);
apply(assumption);
apply(simp);
apply(rule_tac x="fst x" in exI);
apply(rule_tac x="snd x" in exI);
apply(rule conjI);
apply(simp);
apply(rule conjI);
apply(simp);
apply(rule Feas_eq[THEN iffD2]);
apply(simp add: Feas'_def);
apply(rule conjI);
apply(rule notI);
apply(erule notE);
back;
apply(simp add: set_mem_eq);
apply(rule conjI);
apply(rule sublist_subset[THEN iffD2]);
apply(rule subsetI);
apply(simp);
apply(erule disjE);
apply(drule_tac l=B and x=xa in sublistD);
apply(simp);
apply(simp add: set_mem_eq);
apply(drule_tac l=A and x=xa in sublistD);
apply(simp add: set_mem_eq);
apply(simp add: set_mem_eq);
apply(frule_tac l="sortf A" and r="sortf B" in sel_max_F_aux);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(assumption);
apply(frule_tac l="sortf A" and r="sortf B" in sel_max_F_aux2);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(rule conjI);
apply(assumption);
apply(assumption);
apply(frule_tac r="rev(sortf B)" in sel_max_perm);
apply(assumption);
apply(subgoal_tac "sortf A @ [x] <~~> sortf (A @ [x])");
prefer 2;
apply(insert A_sort_date);
apply(drule_tac x="A @ [x]" in spec);
apply(erule conjE);
apply(rule Permutation.trans);
prefer 2;
apply(assumption);
apply(simp);
apply(rule perm_sym);
apply(assumption);
apply(drule_tac ys="sortf (A @ [x])" in perm_sym);
apply(frule_tac ys="sortf A @ [x]" in Permutation.trans);
apply(assumption);
apply(rule F_aux2_sort);
prefer 2;
apply(rule conjI);
apply(rule perm_sym);
apply(assumption);
back;
apply(drule_tac x="A @ [x]" in spec);
apply(erule conjE);
apply(assumption);
apply(rule F_aux25_c);
apply(assumption);
by simp;



section {* Theory morphism *}

thymorph greedy_tau : Greedy --> Jobs
  maps [("Greedy.D" |-> "Jobs.Job"), 
        ("Greedy.V" |-> "nat")]
       [("Greedy.val"      |-> "Jobs.val"),
        ("Greedy.V_ord"    |-> "Jobs.val_ord"),
        ("Greedy.V_op"     |-> "Jobs.val_op"),
        ("Greedy.Feasible" |-> "Jobs.Feas")]





subsection {* Applying the Greedy-Algorithm to the Input and getting the 
           Greedy property by translating the main theorem from the Greedy-Theory *}

translate_thm "Greedy.greedy_main_prop" as "" along greedy_tau;


subsection {* Adapting the main result for this theory *}

constdefs
j_greedy :: "Job list \<Rightarrow> Job list"
"j_greedy js \<equiv> canonical_greedy [] js"


theorem j_greedy_prop : " max_range (j_greedy Input) \<and>
                          (\<forall> x. max_range x \<longrightarrow> Val x \<le> Val (j_greedy Input))"
apply(insert greedy_main_prop);
apply(erule conjE);
apply(rule conjI);
apply(simp add: j_greedy_def);
apply(clarify);
apply(drule_tac x=x in spec);
apply(drule mp, assumption);
apply(simp add: val_ord_def);
by (simp add: j_greedy_def);


end