header {* The exception monad transformer *}

theory ExnT = Main: 

typedecl 'a M

thymorph s : MonadType --> ExnT maps [("MonadType.M" |-> "ExnT.M")] 

(* Get all renamed ops, axioms from monad and redefine the mixfix of bind_ *)
t_instantiate MonadAxms s renames: [("MonadOpEta.eta" |-> "eta_"), 
                                    ("MonadOpBind.bind" |-> "bind_" mixfix: ("_ >>=*\ _" [5, 6] 5) ) ] no_syntax


datatype 'a exn  = Error string | Just 'a

datatype 'a exnT = M "('a exn) M"

(* better, but problem with type synonyms (see STMap):
types 'a exnT = "('a exn) M" 
 *)

consts
 unwrap :: "'a exnT => ('a exn) M"
primrec
 "unwrap (M e) = e"


thymorph t : MonadType --> ExnT
  maps [("MonadType.M" |-> "ExnT.exnT")] 
       

(* Here we need just both consts eta and bind.
   Due to instantiations above no renaming is required 
   and we get the standard mixfix for bind *)
t_instantiate MonadOps t


(* Now we define the operations and prove the monad axioms *)
defs 
 eta_def:  "eta x == M (eta_ (Just x))"
 bind_def: "(v >>= f) == M ((unwrap v) >>=* 
                                  (%exn. case exn of Error e => (eta_ (Error e))
                                                   | Just  x => unwrap (f x)))"

lemma exnT_lunit:  "(eta x >>= t) = t x" 
apply(simp add: bind_def eta_def);
apply(simp add: mon_lunit);
apply(case_tac "t x");
apply(erule ssubst);
by simp;

lemma exnT_runit:  "(t >>= eta) = t"
apply(simp add: bind_def);
apply(case_tac t);
apply(simp);
apply(simp add: eta_def);
apply(rule_tac P="%x. (Ma >>=* x) = Ma" in ssubst);
apply(rule ext);
apply(split exn.split);
apply(rule conjI);
apply(clarify);
apply(rule refl);
by (simp_all add: mon_runit);

lemma exnT_assoc:  "(s >>= t >>= u) = (s >>= (%x. t x >>= u))" 
apply(simp add: bind_def eta_def);
apply(simp add: mon_assoc mon_lunit mon_runit);
apply(rule_tac f="%x. (unwrap s) >>=* x" in arg_cong);
apply(rule ext);
apply(case_tac x);
apply(simp add: mon_lunit);
by simp;


lemma exnT_inj: "eta x= eta y ==> x= y" 
apply(simp add: eta_def);
apply(drule mon_eta_inj);
by simp;


thymorph mon_exnT : MonadAxms --> ExnT
  maps [("MonadType.M" |-> "ExnT.exnT")] 
       [("MonadOpEta.eta" |-> "ExnT.eta"),
        ("MonadOpBind.bind" |-> "ExnT.bind")] 
       [("MonadAxms.mon_lunit" |-> "ExnT.exnT_lunit"),
        ("MonadAxms.mon_runit" |-> "ExnT.exnT_runit"), 
        ("MonadAxms.mon_assoc" |-> "ExnT.exnT_assoc"),
        ("MonadAxms.mon_eta_inj" |-> "ExnT.exnT_inj")]
      strictly

(* Instantiate the exnT monad, but without syntax 
   (too confusing here, as there are too many monads around) *)
t_instantiate Monad mon_exnT

translate_thm "Monad.mon_assoc2" as "" along mon_exnT'

text{* Raising and handling errors *}

constdefs
  raise  :: "string => 'a exnT"
 "raise err == M (eta_ (Error err))" 

constdefs
  handle :: "'a exnT => (string=> 'a M)=> 'a M"
 "handle mex f == (unwrap mex >>=* (% ex. case ex of Error e => f e 
                                                   | Just v  => eta_ v))"
 
end
