header{* The Monad class *}

theory Monad = MonadAxms:

text {* Declares monad syntax *}

constdefs 
  bind' :: "'a M => 'b M => 'b M"             ("_ >>\ _" [5, 6] 5)
  "bind' p q == bind p (% x. q)"
  
text{* This sets up a Haskell-style "do{...}" syntax. *}

nonterminals
  monseq

syntax
  "_monseq"  :: "monseq=> 'a M"                   ("do {(_)}"    [5] 5)
  "_mongen"  :: "[idts, 'a M, monseq]=> monseq"   ("_<-(_);//_"  [10, 6, 5] 5)
  "_monexp"  :: "['a M, monseq]=> monseq"         ("_;//_"       [6, 5] 5)
  "_monexp0" :: "['a M] => monseq"                ("(_)"         5)

translations
  "_monseq(_mongen x a r)" => "a >>= (% x.(_monseq r))"
  "_monseq(_monexp a r)"   => "a >> (_monseq r)"
  "_monseq(_monexp0 a)"    => "a"

print_ast_translation{*

  (* We can't use macro rewrites for the other way around, since
   * this would require an expanding rule a => _monseq(_monexp0 a).
   * Hence we do this as an ast_translation. 
   *
   * The idea is that if we encounter >>= or >>, we traverse along
   * along the right spline to make it into a monseq, and wrap a 
   * _monseq operator around it on the top level. 
   * If we encounter a >>= without a lambda binding afterwards, we leave it
   * alone. 
   *)

  let fun mk_monseq (Appl [Constant "bind", s, 
                           Appl [Constant "_abs", idts, t]]) = 
          Appl[Constant "_mongen", idts, s, mk_monseq t]
        | mk_monseq (Appl [Constant "bind", s, t]) = 
          Appl[Constant "bind", s, mk_monseq t]
        | mk_monseq (Appl [Constant "bind'", s, t]) =
          Appl[Constant "_monexp", s, mk_monseq t]
        (* Base case -- wrap _monexp0 around it *)
        | mk_monseq t = Appl[Constant "_monexp0", t]
       fun mk_mongen (args as [_, Appl (Constant "_abs"::_)]) = 
                Appl[Constant "_monseq", 
                     mk_monseq(Appl (Constant "bind":: args))]
         | mk_mongen _ = raise Match (* not really needed *)
       fun mk_monexp [s, t] = Appl[Constant "_monseq",
                                   mk_monseq(Appl [Constant "bind'", s, t])]
         | mk_monexp _ = raise Match (* mostly occurs if not enough args *)
  in  [("bind", mk_mongen), ("bind'",  mk_monexp)]
  end

 *}

text{* 
  We need associativity in all different flavours with >> instead of >>=.
  We also need a better way of naming these :-)
  All of these lemmas should enable us to always message monadic sequences
  into a right-associative normal form like this: do x<-p1; p2; y <- p3; ... 
  *}

lemma mon_assoc2 [simp]: "(s >>= t >> u) = (do {x<- s; t x; u})"
  by (unfold "bind'_def", simp add: mon_assoc)

lemma mon_assoc4 [simp]: "(do {x<-(do {s; t}); u x}) = (do {s; x<- t; u x})"
  by (unfold bind'_def, simp add: mon_assoc)

lemma mon_assoc5 [simp]: "(do {(do {s; t}); u})= (do {s; t; u})"
  by (unfold bind'_def, simp add: mon_assoc)

constdefs
  mapF  :: "('a => 'b) => 'a M => 'b M"
 "mapF f m == (do {a<-m; eta (f a)})"
  joinF :: " ('a M) M => 'a M "
 "joinF m == (do {a<-m; a})"

lemma mapF1[simp] : " (mapF f) o eta = eta o f " 
apply(rule ext);
by (simp add: mapF_def mon_lunit);

lemmas mapF1'[simp] = mapF1[simplified mapF_def o_def];

lemma mapF2[simp] : " (mapF f) o (mapF g) = mapF (f o g) "
apply(rule ext);
by (simp add: mapF_def mon_assoc mon_lunit);

lemmas mapF2'[simp] = mapF2[simplified o_def mapF_def];

lemma mapF3[simp] : "mapF id = id"
apply(rule ext);
by (simp add: mapF_def mon_runit);

lemma joinF1[simp] : " joinF o eta = id "
apply(rule ext);
by (simp add: joinF_def mon_lunit);

lemmas joinF1'[simp] = joinF1[simplified joinF_def o_def];

lemma mapF_joinF1[simp] : " joinF (mapF f m) = (m >>= f) "
apply(simp add: joinF_def mapF_def);
by (simp add: mon_assoc mon_lunit);

lemmas mapF_joinF1'[simp] = mapF_joinF1[simplified joinF_def mapF_def];

lemma mapF_joinF2[simp] : " joinF o (mapF eta) = id "
apply(rule ext);
apply(simp);
by (rule mon_runit);

lemma mapF_joinF[simp] : " joinF o mapF (mapF f) = (mapF f) o joinF "
apply(rule ext);
by (simp add: mapF_def joinF_def mon_assoc mon_lunit);

lemmas mapF_joinF'[simp] = mapF_joinF[simplified mapF_def joinF_def o_def]; 

end
