header {* An implementation of the state monad as a state transformer *}

theory STMap = Main:

typedecl v

types ref = nat
       S  = "(nat* (nat ~=> v))"
datatype
       'a st = St "S => ('a* S)"

consts 
  unwrap :: "'a st=> (S=> ('a* S))"
primrec
 "unwrap (St s) = s"

lemma wrap: "St (unwrap t) = t"
  by (induct "t", simp)

text{* Get declarations for monad operations: *}
thymorph s: MonadType --> STMap maps [("MonadType.M" |-> "STMap.st")] [] []
t_instantiate MonadOpEta s
t_instantiate MonadOps s' 

(* consts
  eta   :: "'a => 'a st" 
  bind  :: "'a st => ('a => 'b st) => 'b st"  ("_ >>=\ _" [5, 6] 5) *)

defs  
  bind_def: "bind f g == St (split (unwrap o g) o (unwrap f))"
  eta_def:  "eta x    == St (%s. (x, s))"


text{* Proving the monad axioms. *}

lemma mon_lunit:  "(eta x >>= t) = t x"
  by (simp add: bind_def eta_def split_def comp_def wrap) 

lemma mon_runit:  "(t >>= eta)   = t" 
  by (simp add:  bind_def eta_def split_def comp_def wrap)

lemma mon_assoc:  "(s >>= t >>= u) = (s >>= (%x. t x >>= u))" 
  by (simp add: bind_def split_def comp_def wrap)

lemma help: "!!c.(a, c)= (b, c) ==> a = b"
  by (simp add: split_tupled_all)

ML 
{* (* Can't get this to work in Isar: *)
  bind_thm ("help2", read_instantiate_sg (sign_of (theory ("Main"))) 
            [("f", "Pair a"), ("g", "Pair b")] fun_cong RS (thm "help")); 
 *}

lemma mon_eta_inj: "eta x= eta y ==> x= y"
  apply (simp add: eta_def)
  apply (rule help2)
  by simp

text {* The monad instance: *}

thymorph s : MonadAxms --> STMap  maps [("MonadType.M" |-> "STMap.st")] [] []
t_instantiate Monad s

text {* The other operations *}
constdefs
  lookup :: "ref => v st"
 "lookup r == St (%s. (case (snd s) r of None => arbitrary | Some b=> b, s))"

  upd    :: "ref => v => unit st"
 "upd r v  == St (%s. ((), (fst s, (snd s) (r|-> v))))"

  ref    :: "v=> ref st"
 "ref v == St (%s. (fst s+ 1, (fst s+ 1, snd s (fst s|-> v))))"


end
