
header {* The Divide and Conquer scheme *}

theory DaC = Main :

subsection {* Typedecls *}

typedecl d_F 
typedecl d_G 
typedecl r_F 
typedecl r_G


subsection {* Consts *}

consts
 Measure      :: "d_F \<Rightarrow> nat"
 Primitive    :: "d_F \<Rightarrow> bool"
 Dir_Solve    :: "d_F \<Rightarrow> r_F"
 Decompose    :: "d_F \<Rightarrow> (d_G \<times> d_F \<times> d_F)" 
 Compose      :: "(r_G \<times> r_F \<times> r_F) \<Rightarrow> r_F"
 Aux          :: "d_G \<Rightarrow> r_G"
 O_COMPOSE    :: "(r_F \<times> r_G \<times> r_F \<times> r_F) \<Rightarrow> bool"
 O_DECOMPOSE  :: "(d_F \<times> d_G \<times> d_F \<times> d_F) \<Rightarrow> bool"

 F_pre      :: "d_F \<Rightarrow> bool"
 F_post      :: "d_F \<Rightarrow> r_F \<Rightarrow> bool"

 G_pre      :: "d_G \<Rightarrow> bool"
 G_post      :: "d_G \<Rightarrow> r_G \<Rightarrow> bool"




section {* ThyMorph parameters (axioms) and constdefs for them *}

subsection {* Decompose specification *}

constdefs
dcmp_pre :: "d_F \<Rightarrow> bool" 
"dcmp_pre \<equiv> \<lambda> x. F_pre x \<and> (\<not> Primitive x)"


dcmp_post :: " d_F \<Rightarrow> (d_G \<times> d_F \<times> d_F) \<Rightarrow> bool " 
"dcmp_post \<equiv> 
          \<lambda> x (x1, x2, x3). ( G_pre x1 
                               \<and> F_pre x2 
                               \<and> F_pre x3 
                               \<and> ( Measure x2 < Measure x ) 
                               \<and> ( Measure x3 < Measure x )
                               \<and> ( O_DECOMPOSE (x, x1, x2, x3) ) )"



subsection {* Compose specification *}

constdefs
cmps_pre :: "(r_G \<times> r_F \<times> r_F) \<Rightarrow> bool "
" cmps_pre \<equiv> \<lambda>(z1, z2, z3). True "

cmps_post :: "(r_G \<times> r_F \<times> r_F) \<Rightarrow> r_F \<Rightarrow> bool "
" cmps_post \<equiv> \<lambda>(z1, z2, z3) z0. ( O_COMPOSE (z0, z1, z2, z3) ) "



subsection {* Directly-Solve specification *}

constdefs
dsolve_pre :: "d_F \<Rightarrow> bool "
"dsolve_pre \<equiv> \<lambda>x0. ( F_pre x0 \<and> Primitive x0 )"

dsolve_post :: " d_F \<Rightarrow> r_F \<Rightarrow> bool "
"dsolve_post \<equiv> F_post"



subsection {* Strong Problem Reduction Principle (SPRP) *}

constdefs
SPRP :: " (d_F \<times> d_G \<times> d_F \<times> d_F) 
          \<Rightarrow> (r_F \<times> r_G \<times> r_F \<times> r_F) 
          \<Rightarrow> bool "
"SPRP \<equiv> 
     \<lambda>(x0,x1,x2,x3) (z0,z1,z2,z3).( (O_DECOMPOSE (x0, x1, x2, x3)) 
                                     \<and> (G_post x1 z1)
                                     \<and> (F_post x2 z2) 
                                     \<and> (F_post x3 z3) 
                                     \<and> (O_COMPOSE (z0, z1, z2, z3)) )
                                    \<longrightarrow> (F_post x0 z0) "



subsection {* Order on the domain of F *}
constdefs
order_on_df :: "(d_F \<times> d_F)set"
"order_on_df \<equiv> { (x, y). (Measure x) < (Measure y) } "



subsection {* DaC-Axioms *}
axioms
A_DCMP : "\<forall> x. dcmp_pre x \<longrightarrow> dcmp_post x (Decompose x)"
A_AUX  : "\<forall> x. G_pre x \<longrightarrow> G_post x (Aux x)"
A_CMPS : "\<forall> x. cmps_pre x \<longrightarrow> cmps_post x (Compose x)"
A_DS   : "\<forall> x. dsolve_pre x \<longrightarrow> dsolve_post x (Dir_Solve x)"
A_SPRP : "\<forall> x0 x1 x2 x3 z0 z1 z2 z3. 
                    SPRP (x0, x1, x2, x3) (z0, z1, z2, z3)"
A_WF   : "wf order_on_df"



end
