header {* Implementing the Divide-and-Conquer scheme *}

theory DaCimpl = DaC:

constdefs
f_product :: " ('a \<Rightarrow> 'b) \<Rightarrow> ('c \<Rightarrow> 'd) \<Rightarrow> ('e \<Rightarrow> 'f)
               \<Rightarrow> ('a \<times> 'c \<times> 'e) \<Rightarrow> ('b \<times> 'd \<times> 'f) "
"f_product f g h \<equiv> (\<lambda>(a,c,e). (f a, g c, h e))"

  
consts
applyDaC :: " d_F \<Rightarrow> r_F "

recdef applyDaC "measure(\<lambda>x. (Measure x))"
"applyDaC x = (if (Primitive x) then (Dir_Solve x)  
               else (Compose \<circ> (f_product Aux applyDaC applyDaC) \<circ> Decompose) x ) "



subsection {* Divide-and-Conquer main theorem *}


lemma DaC_prim : " Primitive x \<Longrightarrow> applyDaC x = Dir_Solve x " by simp;

lemma DaC_Compose_case : 
"\<lbrakk> 
 \<not>(Primitive x0); 
 Measure (fst (snd (Decompose x0))) < Measure x0; 
 Measure (snd (snd (Decompose x0))) < Measure x0 
 \<rbrakk>  
 \<Longrightarrow> 
  (Compose \<circ> (f_product Aux applyDaC applyDaC) 
           \<circ> Decompose) x0 
  = applyDaC x0"

apply(rule HOL.trans);
prefer 2;
apply(rule HOL.sym);
apply(rule applyDaC.simps);
apply(split split_if);
apply(rule conjI);
apply(simp);
apply(rule impI);
apply(simp only: f_product_def);
apply(simp only: split_def);
apply(simp only: o_def);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(rule refl);
apply(rule impI);
apply(simp only: measure_def inv_image_def);
apply(erule notE);
back;
apply(simp);
apply(rule impI);
apply(simp only: measure_def inv_image_def);
apply(erule notE);
back;
apply(simp);
done




theorem DaC [rule_format]:
   " \<forall> x. F_pre x \<longrightarrow> F_post x (applyDaC x)"
apply(rule allI);
apply(rule_tac r="order_on_df" in wf_induct);
apply(rule A_WF);
apply(clarify);
apply(rule_tac P="%x. F_post xa x" in ssubst);
apply(rule applyDaC.simps);
apply(split split_if);
apply(rule conjI);

-- " Primitive case "
apply(drule_tac x="xa" in spec);
apply(simp add: order_on_df_def del: applyDaC.simps);
apply(rule impI);
apply(insert A_DS);
apply(drule_tac x="xa" in spec);
apply(simp only: dsolve_pre_def);
apply(simp only: dsolve_post_def);
apply(simp);

-- " Non-primitive case "
apply(rule impI);
apply(rename_tac "x0");
apply(insert A_DCMP);
apply(drule_tac x="x0" in spec);
back;
back;
apply(simp add: dcmp_pre_def del: applyDaC.simps);
apply(simp add: dcmp_post_def del: applyDaC.simps);
apply(simp only: split_def);
apply(elim conjE);
apply(frule_tac x="fst(snd (Decompose x0))" in spec);
apply(drule mp);
apply(simp add: order_on_df_def);
apply(drule mp);
apply(assumption);
apply(drule_tac x="snd(snd (Decompose x0))" in spec);
apply(drule mp);
apply(simp add: order_on_df_def);
apply(drule mp);
apply(assumption);
apply(insert A_CMPS);
apply(simp add: cmps_pre_def del: applyDaC.simps);
apply(simp add: cmps_post_def del: applyDaC.simps);
apply(drule_tac x="Aux (fst (Decompose x0))" in spec);
apply(drule_tac x="applyDaC (fst (snd (Decompose x0)))" in spec);
apply(drule_tac x="applyDaC (snd (snd (Decompose x0)))" in spec);
apply(insert A_SPRP);
apply(drule_tac x="x0" in spec);
back;
apply(drule_tac x="fst (Decompose x0)" in spec);
apply(drule_tac x="fst (snd (Decompose x0))" in spec);
back;
apply(drule_tac x="snd (snd (Decompose x0))" in spec);
back;
apply(drule_tac x="applyDaC x0" in spec);
apply(drule_tac x="Aux (fst (Decompose x0))" in spec);
apply(drule_tac x="applyDaC (fst (snd (Decompose x0)))" in spec);
apply(drule_tac x="applyDaC (snd (snd (Decompose x0)))" in spec);
apply(simp add: SPRP_def del: applyDaC.simps);
apply(simp add: split_def del: applyDaC.simps);
apply(drule mp);
apply(rule conjI);
apply(insert A_AUX);
apply(drule_tac x="fst (Decompose x0)" in spec);
apply(drule mp, assumption);
apply(assumption);
apply(drule DaC_Compose_case[simplified o_def f_product_def split_def]);
apply(assumption);
apply(assumption);
apply(erule_tac t="applyDaC x0" in subst);
apply(assumption);
apply(simp (no_asm) only: f_product_def split_def);
apply(frule DaC_Compose_case[simplified o_def f_product_def split_def]);
apply(assumption);
apply(assumption);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(split split_if);
apply(rule conjI);
apply(rule impI);
apply(erule_tac s="applyDaC x0" in ssubst);
apply(assumption);
apply(clarify);
apply(erule notE);
back;
apply(simp (no_asm) add: measure_def inv_image_def);
apply(clarify);
apply(erule notE);
back;
apply(simp (no_asm) add: measure_def inv_image_def);
done

end
