
header {* Insertion-Sort *} 

(* Uses derived constants and theorems from Insertion *)
 

theory InsSort = Insertion + Sorting + Permutation:


subsection {* Assuming the << - relation from Insertion is a linear order *}

axioms
 refl  : " \<forall> x. x << x "
 trans : " \<forall> x y z. x << y \<longrightarrow> y << z \<longrightarrow> x << z "
 total : " \<forall> x y. x << y \<or> x = y \<or> y << x " 


subsection {* Specification of Ins-Sort *}

constdefs

inssort_pre :: "D list \<Rightarrow> bool"
"inssort_pre l \<equiv> True " 

inssort_post :: "D list \<Rightarrow> D list \<Rightarrow> bool"
"inssort_post il ol \<equiv> sorted (op <<) ol \<and> il <~~> ol" 

inssort_aux_pre :: "D \<Rightarrow> bool"
"inssort_aux_pre x \<equiv> True"

inssort_aux_post :: "D \<Rightarrow> D \<Rightarrow> bool"
"inssort_aux_post x y \<equiv> (x = y)"    



subsection {* Implementation of Ins-Sort *}

constdefs

inssort_aux :: " D \<Rightarrow> D "
"inssort_aux \<equiv> id"

inssort_prim :: " D list \<Rightarrow> bool "
"inssort_prim \<equiv> \<lambda>l. (l = [])"

inssort_msr  :: " D list \<Rightarrow> nat "
"inssort_msr \<equiv> length "

inssort_dir_solve :: "D list \<Rightarrow> D list"
"inssort_dir_solve \<equiv> id"

inssort_decompose :: "D list \<Rightarrow> (D \<times> D list \<times> D list)" 
"inssort_decompose \<equiv> \<lambda>l. (hd l, tl l, []) "
 

text {* 
   Here we use the function Insert, which is already derived in Insertion,
   as composition for insertion sort 
*}
constdefs

inssort_compose :: "(D \<times> D list \<times> D list) \<Rightarrow> D list"
"inssort_compose \<equiv> \<lambda>(x, l, r). ins (x, l)"  

inssort_dcmp_rel :: "(D list \<times> D \<times> D list \<times> D list) \<Rightarrow> bool"
"inssort_dcmp_rel \<equiv> \<lambda>(x0, x1, x2, x3). (x0 = (x1#x2)) "

inssort_cmps_rel :: "(D list \<times> D \<times> D list \<times> D list) \<Rightarrow> bool"
"inssort_cmps_rel \<equiv> \<lambda>(z0, z1, z2, z3). ( ins_post (z1, z2) z0 ) "



subsection {* take- and dropWhile properties *}

lemma take_drop_While_Un : " set ( takeWhile P l) \<union> set ( dropWhile P l ) = set l "
apply(induct_tac l);
by auto;

lemma take_While_prop_1[rule_format] : " x mem (takeWhile P l) \<longrightarrow> x mem l "
apply(induct_tac l);
apply(simp);
apply(rule impI);
apply(simp);
apply(rule impI);
apply(split split_if_asm);
apply(simp);
apply(simp);
done

lemma take_While_prop_2 : " set (takeWhile P l) \<subseteq> set l "
apply(rule subsetI);
apply(drule set_mem_eq[THEN iffD2]);
apply(rule set_mem_eq[THEN iffD1]);
by (erule take_While_prop_1);


lemma drop_While_prop_1[rule_format] : " x mem (dropWhile P l) \<longrightarrow> x mem l "
apply(induct_tac l);
apply(simp);
apply(rule impI);
apply(simp);
apply(rule impI);
apply(split split_if_asm);
apply(simp);
apply(simp);
done

lemma drop_While_prop_2 : " set (dropWhile P l) \<subseteq> set l "
apply(rule subsetI);
apply(drule set_mem_eq[THEN iffD2]);
apply(rule set_mem_eq[THEN iffD1]);
by (erule drop_While_prop_1);



subsection {* Sorted property of ins *}

lemma insert_prop[rule_format] : " sorted (op <<) l \<longrightarrow> sorted (op <<) (ins (x, l)) " 
apply(induct_tac l);
apply(simp);
apply(simp add: Insertion.ins_correct [of "(x, [])", simplified ins_pre_def split_def ins_post_def]);
apply(rule impI);
apply(drule mp);
apply(simp);
apply(rule_tac t="ins (x, a # list)" in ssubst);
apply(rule_tac x="(x, a#list)" in Insertion.ins_correct [simplified ins_pre_def split_def ins_post_def]);
apply(simp (no_asm));
apply(simp (no_asm) del: sorted_append sorted.simps);
apply(rule conjI);
apply(rule impI);
apply(unfold gr_def);
apply(erule conjE);
apply(simp only: sorted.simps);
apply(rule conjI);
apply(clarify);
apply(simp only: set_append);
apply(simp only: Un_def);
apply(simp only: mem_Collect_eq);
apply(erule disjE);
apply(drule set_mem_eq[THEN iffD2]);
apply(drule take_While_prop_1);
apply(drule_tac x=y in bspec);
apply(simp add: set_mem_eq);
apply(assumption);
apply(simp only: set.simps);
apply(drule insert_iff[THEN iffD1]);
apply(erule disjE);
apply(clarify);
apply(drule set_mem_eq[THEN iffD2]);
apply(drule drop_While_prop_1);
apply(drule_tac x=y in bspec);
apply(simp add: set_mem_eq);
apply(assumption);
apply(rule_tac P="\<lambda> x. sorted (op <<) x" in ssubst);
apply(rule sym);
apply(rule HOL.trans);
apply(rule_tac x="(x, list)" in Insertion.ins_correct [simplified ins_pre_def split_def ins_post_def]);
apply(simp (no_asm));
apply(simp (no_asm));
apply(assumption);
apply(rule impI);
apply(simp del: sorted.simps);
apply(case_tac "x = a");
apply(clarify);
apply(simp only: sorted.simps);
apply(intro conjI);
apply(clarify);
apply(simp only: set.simps);
apply(drule insert_iff[THEN iffD1]);
apply(erule disjE);
apply(clarify);
apply(simp (no_asm) add: refl);
apply(drule_tac x=y in bspec, assumption);
apply(assumption);
apply(erule conjE);
apply(assumption);
apply(simp (no_asm));
apply(insert total);
apply(drule_tac x=a in spec);
apply(drule_tac x=x in spec);
apply(erule disjE);
apply(drule mp, assumption);
apply(clarify);
apply(erule disjE);
apply(clarify);
apply(rule sorted.simps[THEN iffD2]);
apply(rule conjI);
apply(clarify);
apply(simp only: set.simps);
apply(drule insert_iff[THEN iffD1]);
apply(erule disjE);
apply(clarify);
apply(simp only: sorted.simps);
apply(erule conjE);
apply(drule_tac x=y in bspec, assumption);
apply(rule trans[rule_format]);
by assumption;





subsection {* Local constdefs-"wrappers" of DaC constdefs*}


constdefs
dcmp_pre :: "D list \<Rightarrow> bool" 
"dcmp_pre \<equiv> \<lambda> x. inssort_pre x \<and> (\<not> inssort_prim x) "


dcmp_post :: " D list \<Rightarrow> (D \<times> D list \<times> D list)
                \<Rightarrow> bool " 
"dcmp_post \<equiv> 
               \<lambda> x (x1,x2,x3). ( inssort_aux_pre x1 
                                  \<and> inssort_pre x2 
                                  \<and> inssort_pre x3 
                                  \<and> ( inssort_msr x2 < inssort_msr x ) 
                                  \<and> ( inssort_msr x3 < inssort_msr x )
                                  \<and> ( inssort_dcmp_rel (x, x1, x2, x3)))"



constdefs
cmps_pre :: "(D \<times> D list \<times> D list) \<Rightarrow> bool "
" cmps_pre \<equiv> \<lambda>(z1, z2, z3). True "

cmps_post :: "(D \<times> D list \<times> D list) \<Rightarrow> D list \<Rightarrow> bool "
" cmps_post \<equiv> \<lambda>(z1, z2, z3) z0. ( inssort_cmps_rel (z0, z1, z2, z3) ) "


constdefs
dsolve_pre :: "D list \<Rightarrow> bool "
"dsolve_pre \<equiv> \<lambda>x0. ( inssort_pre x0 \<and> inssort_prim x0 )"

dsolve_post :: " D list \<Rightarrow> D list \<Rightarrow> bool "
"dsolve_post \<equiv> inssort_post"


constdefs
SPRP :: " (D list \<times> D \<times> D list \<times> D list) 
          \<Rightarrow> (D list \<times> D \<times> D list \<times> D list) 
          \<Rightarrow> bool "
"SPRP \<equiv> 
     \<lambda>(x0,x1,x2,x3) (z0,z1,z2,z3).( (inssort_dcmp_rel (x0, x1, x2, x3)) 
                                    \<and> (inssort_aux_post x1 z1)
                                    \<and> (inssort_post x2 z2) 
                                    \<and> (inssort_post x3 z3) 
                                    \<and> (inssort_cmps_rel (z0, z1, z2, z3) )  )
                                    \<longrightarrow> (inssort_post x0 z0)  "


constdefs
order_on_df :: "(D list \<times> D list) set"
"order_on_df \<equiv> { (x, y). (inssort_msr x) < (inssort_msr y) } "



subsection {* Proof of the DaC-axioms *}



-- " decompose "
lemma a_dcmp : "\<forall> x. dcmp_pre x \<longrightarrow> dcmp_post x (inssort_decompose x)"
apply(clarify);
apply(simp add: dcmp_pre_def dcmp_post_def inssort_prim_def inssort_msr_def inssort_decompose_def);
apply(simp add: inssort_pre_def inssort_aux_pre_def inssort_dcmp_rel_def );
done

-- " aux "
lemma a_aux : "\<forall> x. inssort_aux_pre x \<longrightarrow> inssort_aux_post x (inssort_aux x)"
apply(clarify);
apply(simp add: inssort_aux_pre_def inssort_aux_post_def inssort_aux_def);
done


-- " dsolve "
lemma a_ds :  "\<forall> x. dsolve_pre x \<longrightarrow> dsolve_post x (inssort_dir_solve x)"
apply(clarify);
apply(simp add: dsolve_pre_def dsolve_post_def inssort_dir_solve_def);
apply(simp add: inssort_prim_def inssort_pre_def inssort_post_def);
done


-- " wf "
lemma a_wf : "wf order_on_df" 
apply(simp add: order_on_df_def inssort_msr_def);
apply(rule_tac P="wf" in ssubst);
apply(subgoal_tac "{(x, y). (length x < length y) } = {(x, y). ((length x, length y) \<in> less_than) }");
apply(assumption);
apply(simp);
apply(fold inv_image_def);
apply(rule wf_inv_image);
apply(simp);
done


-- " compose "
lemma a_cmps : "\<forall> x. cmps_pre x \<longrightarrow> cmps_post x (inssort_compose x)"
apply(clarify);
apply(simp add: cmps_pre_def cmps_post_def inssort_cmps_rel_def inssort_compose_def);
apply(rule Insertion.ins_correct);
by (simp add: ins_pre_def);



-- " SPRP "
lemma a_sprp : "\<forall> x0 x1 x2 x3 z0 z1 z2 z3. 
                    SPRP (x0, (x1, x2, x3)) (z0, (z1, z2, z3))"
apply(intro allI);
apply(simp add: SPRP_def);
apply(clarify);
apply(simp add: inssort_aux_post_def inssort_post_def inssort_cmps_rel_def inssort_dcmp_rel_def);
apply(elim conjE);
apply(rule conjI);
apply(drule_tac x="z1" in insert_prop);
apply(subgoal_tac "ins (z1, z2) = z0");
apply(erule_tac t=z0 in subst);
apply(assumption);
apply(rule HOL.trans);
apply(rule Insertion.ins_correct[simplified ins_pre_def split_def ins_post_def]);
apply(simp (no_asm));
apply(simp (no_asm));
apply(simp add: ins_post_def);
apply(simp add: ins_post_def);
apply(rule Permutation.trans);
prefer 2;
apply(rule perm_append_Cons);
apply(rule Permutation.Cons);
by simp;




subsubsection {* Instantiating divide-and-conquer for insertion sort *}

thymorph tau_ins_sort : DaC --> InsSort 
         maps [("DaC.d_F" |-> "D list"),
               ("DaC.d_G" |-> "Insertion.D"),
               ("DaC.r_F" |-> "D list"),
               ("DaC.r_G" |-> "Insertion.D")]
              [("DaC.Primitive"   |-> "InsSort.inssort_prim"),
               ("DaC.O_DECOMPOSE" |-> "InsSort.inssort_dcmp_rel"),
               ("DaC.O_COMPOSE"   |-> "InsSort.inssort_cmps_rel"), 
               ("DaC.G_pre"       |-> "InsSort.inssort_aux_pre"),  
               ("DaC.G_post"      |-> "InsSort.inssort_aux_post"),
               ("DaC.F_pre"       |-> "InsSort.inssort_pre"), 
               ("DaC.F_post"      |-> "InsSort.inssort_post"),
               ("DaC.Compose"     |-> "InsSort.inssort_compose"), 
               ("DaC.Aux"         |-> "InsSort.inssort_aux") ]
             

t_instantiate DaCimpl tau_ins_sort renames: [("DaCimpl.applyDaC" |-> "isort")]

translate_thm "DaCimpl.DaC" as isort_corr along tau_ins_sort'

text {* Correctness of insertion sort, pre/post unfolded. *}

theorem isort_correct : "sorted (op <<) (isort l) \<and> l <~~> (isort l) "
  by (insert isort_corr, simp add: inssort_pre_def inssort_post_def)


end
