
header {* Insertion *}

theory Insertion = Trafo :


subsection {* Parameters of Insertion: domain and binary relation *}

typedecl D

consts
le :: " D \<Rightarrow> D \<Rightarrow> bool "   (infixl "<<" 45)

subsection {* Specification of Insertion *}

types D_I = "D \<times> D list"


constdefs
gr :: "D \<Rightarrow> D \<Rightarrow> bool"
"gr x y \<equiv> (y << x) \<and> (x \<noteq> y)"


ins_pre :: "D_I \<Rightarrow> bool "
"ins_pre \<equiv> \<lambda>(a, l). True "

ins_post :: "D_I \<Rightarrow> D list \<Rightarrow> bool"
"ins_post \<equiv> \<lambda>(x, l) lt. ( lt = ( (takeWhile (gr x) l) @[x]@ (dropWhile (gr x) l) ) ) "       

aux_pre :: "D \<Rightarrow> bool"
"aux_pre \<equiv> \<lambda>x. True"

aux_post :: "D \<Rightarrow> D \<Rightarrow> bool"
"aux_post \<equiv> \<lambda>x y. ( x = y )"



subsection {* Implementation of insertion *}

constdefs

ins_msr  :: "D_I \<Rightarrow> nat"
"ins_msr \<equiv> \<lambda>(x, l). length l"

ins_aux  :: "D \<Rightarrow> D"
"ins_aux \<equiv> id"

ins_prim :: " D_I \<Rightarrow> bool"
"ins_prim \<equiv> \<lambda>(x, l). (l = []) \<or> (\<not> gr x (hd l))"

ins_dir_solve :: "D_I \<Rightarrow> D list"
"ins_dir_solve \<equiv> \<lambda>(x, l). (x#l)"

ins_decompose :: "D_I \<Rightarrow> (D \<times> D_I \<times> D_I)" 
"ins_decompose \<equiv> \<lambda>(x, l). (hd l, (x, tl l), (x, []))"
 
ins_compose :: "(D \<times> D list \<times> D list) \<Rightarrow> D list"
"ins_compose \<equiv> \<lambda>(x, l, r). (x#l)"

ins_dcmp_rel :: "(D_I \<times> D \<times> D_I \<times> D_I) \<Rightarrow> bool"
"ins_dcmp_rel \<equiv> \<lambda>(x0, x1, x2, x3). (x0 = (fst x2, (x1# snd x2))) \<and> (gr (fst x2) x1) "

ins_cmps_rel :: "(D list \<times> D \<times> D list \<times> D list) \<Rightarrow> bool"
"ins_cmps_rel \<equiv> \<lambda>(z0, z1, z2, z3). (z0 = z1#z2) "

 
subsection {* Local constdefs-"wrappers" of DaC constdefs*}

constdefs
dcmp_pre :: "D_I \<Rightarrow> bool" 
"dcmp_pre \<equiv> \<lambda>x0. (ins_pre x0 \<and> (\<not> ins_prim x0) )"


dcmp_post :: " D_I \<Rightarrow> (D \<times> D_I \<times> D_I)
               \<Rightarrow> bool " 
"dcmp_post \<equiv> 
          \<lambda>x0 (x1,x2,x3). ( aux_pre x1 
                             \<and> ins_pre x2 
                             \<and> ins_pre x3 
                             \<and> ( ins_msr x2 < ins_msr x0 ) 
                             \<and> ( ins_msr x3 < ins_msr x0 )
                             \<and> ( ins_dcmp_rel (x0, x1, x2, x3)))"



constdefs
cmps_pre :: "(D \<times> D list \<times> D list) \<Rightarrow> bool "
"cmps_pre \<equiv> \<lambda>(z1, z2, z3). True "

cmps_post :: "(D \<times> D list \<times> D list) 
              \<Rightarrow> D list \<Rightarrow> bool "
"cmps_post \<equiv> \<lambda>(z1, z2, z3) z0. ( ins_cmps_rel (z0, z1, z2, z3) ) "


constdefs
dsolve_pre :: "D_I \<Rightarrow> bool "
"dsolve_pre \<equiv> \<lambda>x0. ( ins_pre x0 \<and> ins_prim x0 )"

dsolve_post :: " D_I \<Rightarrow> D list \<Rightarrow> bool "
"dsolve_post \<equiv> ins_post"


constdefs
SPRP :: " (D_I \<times> D \<times> D_I \<times> D_I) 
          \<Rightarrow> (D list \<times> D \<times> D list \<times> D list) 
          \<Rightarrow> bool "
"SPRP \<equiv> 
     \<lambda>(x0,x1,x2,x3) (z0,z1,z2,z3).( (ins_dcmp_rel (x0, x1, x2, x3)) 
                                    \<and> (aux_post x1 z1)
                                    \<and> (ins_post x2 z2) 
                                    \<and> (ins_post x3 z3) 
                                    \<and> (ins_cmps_rel (z0, z1, z2, z3) )  )
                                    \<longrightarrow> (ins_post x0 z0)  "


constdefs
order_on_df :: "(D_I \<times> D_I)set"
"order_on_df \<equiv> { (x, y). (ins_msr x) < (ins_msr y) } "


subsection {* Proof of the DaC-axioms *}


-- A_DCMP
lemma a_dcmp : "\<forall> x. dcmp_pre x \<longrightarrow> dcmp_post x (ins_decompose x)"
apply(clarify);
apply(simp add: dcmp_pre_def dcmp_post_def);
apply(simp add: ins_decompose_def ins_prim_def ins_msr_def);
apply(simp add: aux_pre_def ins_pre_def ins_post_def);
apply(simp add: ins_dcmp_rel_def);
done




-- A_AUX
lemma a_aux :  "\<forall> x. aux_pre x \<longrightarrow> aux_post x (ins_aux x)"
by (simp add: aux_pre_def aux_post_def ins_aux_def);

-- A_CMPS
lemma a_cmps :  "\<forall> x. cmps_pre x \<longrightarrow> cmps_post x (ins_compose x)"
apply(clarify);
apply (simp add: cmps_pre_def cmps_post_def);
by (simp add: ins_compose_def ins_cmps_rel_def);

-- A_DS
lemma a_ds : "\<forall> x. dsolve_pre x \<longrightarrow> dsolve_post x (ins_dir_solve x)"
apply(clarify);
apply (simp add: dsolve_pre_def dsolve_post_def);
apply (simp add: ins_pre_def ins_prim_def ins_post_def);
apply(simp add: ins_dir_solve_def);
apply(case_tac b);
apply(simp);
apply(erule disjE);
apply(simp);
apply(clarify);
by simp;


-- A_SPRP
lemma a_sprp : "\<forall> x0 x1 x2 x3 z0 z1 z2 z3. 
                    SPRP (x0, (x1, x2, x3)) (z0, (z1, z2, z3))"

apply(intro allI);
apply(simp add: SPRP_def);
apply(rule impI);
apply(elim conjE);
apply(simp add: ins_dcmp_rel_def ins_cmps_rel_def);
apply(elim conjE);
apply(simp add: aux_post_def ins_pre_def ins_post_def split_def);
by simp;

--  A_WF 
lemma a_wf : "wf order_on_df"
apply(simp add: order_on_df_def ins_msr_def);
apply(rule_tac P="wf" in ssubst);
apply(subgoal_tac "{(x, y). split (\<lambda> x. size) x < split (\<lambda> x. size) y } 
                 = {(x, y). ((split (\<lambda> x. size) x, split (\<lambda> x. size) y) \<in> less_than) }");
apply(assumption);
apply(simp);
apply(fold inv_image_def);
apply(rule wf_inv_image);
by simp;



subsection {* Instantiating divide-and-conquer *}


thymorph tau_ins : DaC --> Insertion
         maps [("DaC.d_F" |-> "D_I"), 
               ("DaC.d_G" |-> "Insertion.D"),
               ("DaC.r_F" |-> "D list"),
               ("DaC.r_G" |-> "Insertion.D")] 
              [("DaC.Primitive"    |->     "Insertion.ins_prim"), 
               ("DaC.F_pre"        |->     "Insertion.ins_pre"),
               ("DaC.F_post"       |->     "Insertion.ins_post"),
               ("DaC.G_post"       |->     "Insertion.aux_post"),
               ("DaC.Aux"          |->     "Insertion.ins_aux") ]
             


t_instantiate DaCimpl tau_ins renames: [("DaCimpl.applyDaC" |-> "ins")]

translate_thm "DaCimpl.DaC" as ins_correct along tau_ins'

end
