
header {* QuickSort *}


theory QuickSort = Trafo + Permutation + Sorting :


subsection {* Parameters of QuickSort: domain and order *}

typedecl D

consts
le :: " D \<Rightarrow> D \<Rightarrow> bool "   (infixl "<<" 45)

axioms
 refl  : " \<forall> x. x << x "
 trans : " \<forall> x y z. x << y \<longrightarrow> y << z \<longrightarrow> x << z "
 total : " \<forall> x y. x << y \<or> x = y \<or> y << x " 




subsection {* Specification of QuickSort *}

constdefs     
qsort_pre :: "D list \<Rightarrow> bool"
"qsort_pre \<equiv> (\<lambda>x. True)"

qsort_post :: " D list \<Rightarrow> D list \<Rightarrow> bool " 
"qsort_post \<equiv> (\<lambda>il ol. sorted (op <<) ol \<and> il <~~> ol) " 


qsort_aux_pre :: "D \<Rightarrow> bool"
"qsort_aux_pre \<equiv> (\<lambda>x. True)"  

qsort_aux_post :: "D \<Rightarrow> D \<Rightarrow> bool"
"qsort_aux_post \<equiv> (\<lambda>x y. x=y) " 




subsection {* Implementation of QuickSort *}

subsubsection {* Basic functions for Quicksort_Decomposition *}

constdefs
low :: " D \<Rightarrow> D list \<Rightarrow> D list "
"low x xs \<equiv> filter (\<lambda>y. y << x) xs"

high :: " D \<Rightarrow> D list \<Rightarrow> D list "
"high x xs \<equiv> filter (\<lambda>y. (\<not> (y << x))) xs"


subsubsection {* Permutation properties of low and high *}

lemma low_1a[rule_format] : " a << x \<longrightarrow> low x (a#xs) = a#(low x xs) "
apply(induct_tac xs);
apply(simp add: low_def);
apply(rule impI);
apply(drule mp, assumption);
by(simp add: low_def);

lemma low_1b[rule_format] : " \<not>(a << x) \<longrightarrow> low x (a#xs) = low x xs "
apply(induct_tac xs);
apply(simp add: low_def);
apply(rule impI);
apply(drule mp, assumption);
by(simp add: low_def);

lemma high_1a[rule_format] : " a << x \<longrightarrow> high x (a#xs) = high x xs "
apply(induct_tac xs);
apply(simp add: high_def);
apply(rule impI);
apply(drule mp, assumption);
by(simp add: high_def);

lemma high_1b[rule_format] : " \<not>(a << x) \<longrightarrow> high x (a#xs) = a#(high x xs) "
apply(induct_tac xs);
apply(simp add: high_def);
apply(rule impI);
apply(drule mp, assumption);
by(simp add: high_def);


lemma low_high_perm : " (low x xs) @ [x] @ (high x xs) <~~> (x#xs) "
apply(induct_tac xs);
apply(simp add: low_def high_def);
apply(case_tac "a << x");
apply(frule_tac xs="list" in low_1a);
apply(erule ssubst);
apply(drule_tac xs="list" in high_1a);
apply(erule ssubst);
apply(rule Permutation.trans);
prefer 2;
apply(rule Permutation.swap);
apply(simp);
apply(frule_tac xs="list" in low_1b);
apply(erule ssubst);
apply(drule_tac xs="list" in high_1b);
apply(erule ssubst);
apply(rule_tac ys="a#(low x list @ (x#(high x list)))" in Permutation.trans);
apply(rule perm_sym);
apply(rule Permutation.trans);
apply(rule_tac a="a" and ys="x # (high x list)" in perm_append_Cons);
apply(simp);
apply(rule Permutation.trans);
apply(rule Permutation.swap);
apply(rule perm_refl);
apply(rule Permutation.trans);
prefer 2;
apply(rule Permutation.swap);
apply(simp);
done



subsubsection {* Measure properties of low and high *}

lemma le_low : " length (low x xs) \<le> length xs " by (simp add: low_def);

lemma le_high : " length (high x xs) \<le> length xs " by (simp add: high_def);

lemma less_low : " length (low x l) < length (x#l) "
apply(simp);
apply(rule less_Suc_eq_le[THEN iffD2]);
by (rule le_low);

lemma less_high : " length (high x l) < length (x#l) "
apply(simp);
apply(rule less_Suc_eq_le[THEN iffD2]);
by (rule le_high);


subsubsection {* DaC-Scheme for Quicksort *}

constdefs
qsort_aux  :: " D \<Rightarrow> D"
"qsort_aux \<equiv> id"

qsort_prim :: " (D list) \<Rightarrow> bool  "
"qsort_prim \<equiv> \<lambda>l. (l = [])"

qsort_msr  :: " (D list) \<Rightarrow> nat "
"qsort_msr \<equiv> length "

qsort_dir_solve :: "D list \<Rightarrow> D list"
"qsort_dir_solve \<equiv> id"

qsort_decompose :: "D list \<Rightarrow> (D \<times> D list \<times> D list)" 
"qsort_decompose \<equiv> \<lambda>l. (hd l, low (hd l) (tl l), (high (hd l) (tl l))) "
 
qsort_compose :: "(D \<times> D list \<times> D list) \<Rightarrow> D list "
"qsort_compose \<equiv> \<lambda>(x, l, r). (l @ [x] @ r)"

qsort_dcmp_rel :: "(D list \<times> D \<times> D list \<times> D list) \<Rightarrow> bool"
"qsort_dcmp_rel \<equiv> \<lambda>(x0, x1, x2, x3). (x0 <~~> (x2 @ [x1] @ x3)) \<and>
                          (\<forall> x. x \<in> set x2 \<longrightarrow> x << x1) \<and> 
                          (\<forall> x. x \<in> set x3 \<longrightarrow> (\<not> x << x1)) "

qsort_cmps_rel :: "(D list \<times> D \<times> D list \<times> D list) \<Rightarrow> bool"
"qsort_cmps_rel \<equiv> \<lambda>(z0, z1, z2, z3). ( z0 = (z2 @ [z1] @ z3) ) "




subsection {* Local constdefs-"wrappers" of DaC constdefs*}

constdefs
dcmp_pre :: "D list \<Rightarrow> bool" 
"dcmp_pre \<equiv> \<lambda> x. qsort_pre x \<and> (\<not> qsort_prim x)"


dcmp_post :: " D list \<Rightarrow> (D \<times> D list \<times> D list)
                \<Rightarrow> bool " 
"dcmp_post \<equiv> 
          \<lambda> x (x1,x2,x3). (   qsort_aux_pre x1 
                               \<and> qsort_pre x2 
                               \<and> qsort_pre x3 
                               \<and> ( qsort_msr x2 < qsort_msr x ) 
                               \<and> ( qsort_msr x3 < qsort_msr x )
                               \<and> ( qsort_dcmp_rel (x, x1, x2, x3)))"



constdefs
cmps_pre :: "(D \<times> D list \<times> D list) \<Rightarrow> bool "
"cmps_pre \<equiv> \<lambda>(z1, z2, z3). True "

cmps_post :: "(D \<times> D list \<times> D list) \<Rightarrow> D list \<Rightarrow> bool "
"cmps_post \<equiv> \<lambda>(z1, z2, z3) z0. ( qsort_cmps_rel (z0, z1, z2, z3) ) "


constdefs
dsolve_pre :: "D list \<Rightarrow> bool "
"dsolve_pre \<equiv> \<lambda>x0. ( qsort_pre x0 \<and> qsort_prim x0 )"


dsolve_post :: " D list \<Rightarrow> D list \<Rightarrow> bool "
"dsolve_post \<equiv> qsort_post"


constdefs
SPRP :: " (D list \<times> D \<times> D list \<times> D list) 
          \<Rightarrow> (D list \<times> D \<times> D list \<times> D list) 
          \<Rightarrow> bool "
"SPRP \<equiv> 
 \<lambda>(x0,x1,x2,x3) (z0,z1,z2,z3).( (qsort_dcmp_rel (x0, x1, x2, x3)) 
                                 \<and> (qsort_aux_post x1 z1)
                                 \<and> (qsort_post x2 z2) 
                                 \<and> (qsort_post x3 z3) 
                                 \<and> (qsort_cmps_rel (z0, z1, z2, z3) )  )
                      \<longrightarrow> (qsort_post x0 z0)  "


constdefs
order_on_df :: "(D list \<times> D list)set"
"order_on_df \<equiv> { (x, y). (qsort_msr x) < (qsort_msr y) } "





subsection {* Proof of the DaC axioms *}



-- " proof of A_DCMP "
lemma a_dcmp : "\<forall> x. dcmp_pre x \<longrightarrow> dcmp_post x (qsort_decompose x)"
apply(clarify);
apply(simp add: dcmp_pre_def dcmp_post_def);
apply(simp add: qsort_decompose_def);
apply(simp add: qsort_dcmp_rel_def); 
apply(simp add: qsort_pre_def qsort_prim_def qsort_msr_def);
apply(drule hd_Cons_tl);
apply(intro conjI);
apply(simp add: qsort_aux_pre_def);
apply(erule subst, rule less_low);
apply(erule subst, rule less_high);
apply(erule subst, rule low_high_perm[simplified, THEN perm_sym]);
apply(clarify);
apply(simp add: low_def);
apply(clarify);
by (simp add: high_def);


-- " proof of A_AUX "
lemma a_aux : "\<forall> x. qsort_aux_pre x \<longrightarrow> qsort_aux_post x (qsort_aux x)"
by (simp add: qsort_aux_pre_def qsort_aux_post_def qsort_aux_def);



-- " proof of A_CMPS "
lemma a_cmps : "\<forall> x. cmps_pre x \<longrightarrow> cmps_post x (qsort_compose x)"
apply(clarify);
apply(simp add: cmps_pre_def cmps_post_def);
by (simp add: qsort_cmps_rel_def qsort_compose_def);



-- " proof of A_DS "
lemma a_ds :  "\<forall> x. dsolve_pre x \<longrightarrow> dsolve_post x (qsort_dir_solve x)"
apply(clarify);
apply(simp add: dsolve_pre_def dsolve_post_def qsort_dir_solve_def);
apply(simp add: qsort_pre_def qsort_post_def qsort_prim_def);
done




-- " proof of A_SPRP "
lemma a_sprp : "\<forall> x0 x1 x2 x3 z0 z1 z2 z3. 
                    SPRP (x0, x1, x2, x3) (z0, z1, z2, z3)"
apply(simp add: SPRP_def);
apply(clarify);
apply(simp add: qsort_dcmp_rel_def qsort_cmps_rel_def);
apply(simp add: qsort_post_def qsort_aux_post_def);
apply(elim conjE);
apply(rule conjI);
apply(clarify);
apply(drule_tac xs="x3" in perm_sym);
apply(drule_tac xs="z3" and x=x in perm_mem);
apply(simp add: set_mem_eq);
apply(drule_tac x=x in spec);
back;
apply(drule mp, simp add: set_mem_eq);
apply(insert total);
apply(drule_tac x=x in spec);
back;
apply(drule_tac x=z1 in spec);
back;
apply(erule disjE);
apply(clarify);
apply(erule disjE);
apply(clarify);
apply(simp (no_asm) add: refl);
apply(assumption);
apply(rule conjI);
apply(clarify);
apply(rule conjI);
apply(drule_tac xs="x2" in perm_sym);
apply(drule_tac xs="z2" and x=x in perm_mem);
apply(simp add: set_mem_eq);
apply(drule_tac x=x in spec);
apply(drule mp, simp add: set_mem_eq);
apply(assumption);
apply(clarify);
apply(drule_tac xs="x2" in perm_sym);
apply(drule_tac xs="x3" in perm_sym);
apply(drule_tac xs="z2" and x=x in perm_mem);
apply(simp add: set_mem_eq);
apply(drule_tac xs="z3" and x=xa in perm_mem);
apply(simp add: set_mem_eq);
apply(drule_tac x=x in spec);
apply(drule mp, simp add: set_mem_eq);
apply(drule_tac x=xa in spec);
apply(drule mp, simp add: set_mem_eq);
apply(drule_tac x=xa in spec);
apply(drule_tac x=z1 in spec);
apply(erule disjE);
apply(clarify);
apply(erule disjE);
apply(clarify);
apply(erule QuickSort.trans[rule_format]);
apply(assumption);
apply(rule Permutation.trans, assumption);
apply(rule Permutation.trans);
apply(rule perm_sym);
apply(rule perm_append_Cons);
apply(rule Permutation.trans);
prefer 2;
apply(rule perm_append_Cons);
apply(clarify);
apply(drule_tac l="x3" and xs="x2" in perm_append2);
apply(rule Permutation.trans, assumption);
apply(rule perm_append1);
by assumption;




-- " proof of A_WF "
lemma a_wf : "wf order_on_df" 
apply(simp add: order_on_df_def qsort_msr_def);
apply(rule_tac P="wf" in ssubst);
apply(subgoal_tac "{(x, y). (length x < length y) } = {(x, y). ((length x, length y) \<in> less_than) }");
apply(assumption);
apply(simp);
apply(fold inv_image_def);
apply(rule wf_inv_image);
by simp;


subsection {* We use a theory morphism to prove correctness *}

thymorph tau_qsort : DaC --> QuickSort
         maps [("DaC.d_F" |-> "D list"), 
               ("DaC.d_G" |-> "QuickSort.D"),
               ("DaC.r_F" |-> "D list"),
               ("DaC.r_G" |-> "QuickSort.D")] 
              [("DaC.Primitive"    |->     "QuickSort.qsort_prim"),
               ("DaC.O_DECOMPOSE"  |->     "QuickSort.qsort_dcmp_rel"),
               ("DaC.O_COMPOSE"    |->     "QuickSort.qsort_cmps_rel"),
               ("DaC.F_pre"        |->     "QuickSort.qsort_pre"), 
               ("DaC.F_post"       |->     "QuickSort.qsort_post"),
               ("DaC.Aux"          |->     "QuickSort.qsort_aux"),
               ("DaC.G_post"       |->     "QuickSort.qsort_aux_post")]
             

(* Instantiate the divide-and-conquer scheme *)

t_instantiate DaCimpl tau_qsort 
  renames: [("DaCimpl.applyDaC" |-> "quicksort")]

(* Translate the main result *)
translate_thm "DaCimpl.DaC" as quicksort_corr along tau_qsort'

subsection {* The correctness result, with pre&post-condition unfolded *}

theorem quicksort_correct: "sorted (op <<) (quicksort l) \<and> l <~~> (quicksort l)"
 by (insert quicksort_corr, simp add: qsort_pre_def qsort_post_def);

end
